import pytest

from vizro.tables.dash_table import _set_defaults_nested


@pytest.fixture
def default_dictionary():
    return {"a": {"b": {"c": 1, "d": 2}}, "e": 3}


@pytest.mark.parametrize(
    "input, expected",
    [
        ({}, {"a": {"b": {"c": 1, "d": 2}}, "e": 3}),  # nothing supplied
        ({"e": 10}, {"a": {"b": {"c": 1, "d": 2}}, "e": 10}),  # flat main key
        ({"a": {"b": {"c": 11, "d": 12}}}, {"a": {"b": {"c": 11, "d": 12}}, "e": 3}),  # updated multiple nested keys
        ({"a": {"b": {"c": 1, "d": {"f": 42}}}}, {"a": {"b": {"c": 1, "d": {"f": 42}}}, "e": 3}),  # add new dict
        ({"a": {"b": {"c": 5}}}, {"a": {"b": {"c": 5, "d": 2}}, "e": 3}),  # arbitrary nesting
    ],
)
def test_set_defaults_nested(default_dictionary, input, expected):
    assert _set_defaults_nested(input, default_dictionary) == expected



"""File to store constants."""
ALL_OPTION = "ALL"
NONE_OPTION = "NONE"
STATIC_URL_PREFIX = "vizro"
MODULE_PAGE_404 = "not_found_404"
EMPTY_SPACE_CONST = -1
ON_PAGE_LOAD_ACTION_PREFIX = "on_page_load_action"
FILTER_ACTION_PREFIX = "filter_action"
PARAMETER_ACTION_PREFIX = "parameter_action"
ACCORDION_DEFAULT_TITLE = "SELECT PAGE"



# ruff: noqa: F403, F405
import os
import runpy
from pathlib import Path

import chromedriver_autoinstaller_fix
import pytest

from vizro import Vizro


# Use monkeypatch as a session-scoped fixture.
# Taken from https://github.com/pytest-dev/pytest/issues/363#issuecomment-1335631998.
@pytest.fixture(scope="session")
def monkeypatch_session():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(scope="session", autouse=True)
def setup_integration_test_environment(monkeypatch_session):
    # Dash debug mode seems to interfere with the tests, so we disable it here. Note "false" as a string is correct.
    monkeypatch_session.setenv("DASH_DEBUG", "false")
    # We only need to install chromedriver outside CI.
    if not os.getenv("CI"):
        chromedriver_autoinstaller_fix.install()


@pytest.fixture(params=["default", "from_dict", "from_json", "from_yaml"])
def dashboard(request, monkeypatch):
    monkeypatch.chdir(Path(__file__).parents[2] / f"examples/{request.param}")
    app = runpy.run_path("app.py")
    return app["dashboard"]


# Ignore deprecation warning until this is solved: https://github.com/plotly/dash/issues/2590
@pytest.mark.filterwarnings("ignore:HTTPResponse.getheader()")
def test_dashboard(dash_duo, dashboard):
    app = Vizro(assets_folder=Path(__file__).parents[2] / "examples/assets").build(dashboard).dash
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



"""Unit tests for vizro.models.Action."""

import json
import sys

import pandas as pd
import plotly
import pytest
from dash import Output, State, html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro
from vizro.actions import export_data
from vizro.managers import model_manager
from vizro.models._action._action import Action
from vizro.models.types import capture


@pytest.fixture
def custom_action_function_mock_return(request):
    @capture("action")
    def _custom_action_function_mock_return():
        return request.param

    return _custom_action_function_mock_return


@pytest.fixture
def custom_action_build_expected():
    return html.Div(
        children=[],
        id="action_test_action_model_components_div",
        hidden=True,
    )


@pytest.fixture
def predefined_action_build_expected():
    return html.Div(children=[], id="filter_action_test_filter_action_model_components_div", hidden=True)


class TestActionInstantiation:
    """Tests model instantiation."""

    def test_create_action_mandatory_only(self, identity_action_function):
        function = identity_action_function()
        action = Action(function=function)

        assert hasattr(action, "id")
        assert action.function is function
        assert action.inputs == []
        assert action.outputs == []

    def test_create_action_mandatory_and_optional(self, identity_action_function):
        function = identity_action_function()
        inputs = ["component_1.property_A", "component_1.property_B"]
        outputs = ["component_2.property_A", "component_2.property_B"]

        action = Action(function=function, inputs=inputs, outputs=outputs)

        assert hasattr(action, "id")
        assert action.function is function
        assert action.inputs == inputs
        assert action.outputs == outputs

    @pytest.mark.parametrize(
        "inputs, outputs",
        [
            ([], []),
            (["component.property"], ["component.property"]),
            (["component.property", "component.property"], ["component.property", "component.property"]),
            (
                ["Component_1.Property_A", "Component_2.Property_B"],
                ["Component_1.Property_A", "Component_2.Property_B"],
            ),
        ],
    )
    def test_inputs_outputs_valid(self, inputs, outputs, identity_action_function):
        action = Action(function=identity_action_function(), inputs=inputs, outputs=outputs)

        assert action.inputs == inputs
        assert action.outputs == outputs

    @pytest.mark.parametrize(
        "inputs",
        [
            [""],
            ["component"],
            ["component_property"],
            ["component.property.property"],
        ],
    )
    def test_inputs_invalid(self, inputs, identity_action_function):
        with pytest.raises(ValidationError, match="string does not match regex"):
            Action(function=identity_action_function(), inputs=inputs, outputs=[])

    @pytest.mark.parametrize(
        "outputs",
        [
            [""],
            ["component"],
            ["component_property"],
            ["component.property.property"],
        ],
    )
    def test_outputs_invalid(self, outputs, identity_action_function):
        with pytest.raises(ValidationError, match="string does not match regex"):
            Action(function=identity_action_function(), inputs=[], outputs=outputs)

    @pytest.mark.parametrize("file_format", [None, "csv", "xlsx"])
    def test_export_data_file_format_valid(self, file_format):
        action = Action(id="action_test", function=export_data(file_format=file_format))
        assert action.id == "action_test"
        assert action.inputs == []
        assert action.outputs == []

    def test_export_data_file_format_invalid(self):
        with pytest.raises(
            ValueError, match='Unknown "file_format": invalid_file_format.' ' Known file formats: "csv", "xlsx".'
        ):
            Action(function=export_data(file_format="invalid_file_format"))

    def test_export_data_xlsx_without_required_libs_installed(self, monkeypatch):
        monkeypatch.setitem(sys.modules, "openpyxl", None)
        monkeypatch.setitem(sys.modules, "xlswriter", None)

        with pytest.raises(
            ModuleNotFoundError, match="You must install either openpyxl or xlsxwriter to export to xlsx format."
        ):
            Action(function=export_data(file_format="xlsx"))


@pytest.fixture
def managers_one_page_without_graphs_one_button():
    """Instantiates a simple model_manager and data_manager with a page, and no graphs."""
    vm.Page(
        id="test_page",
        title="Test page",
        components=[vm.Graph(figure=px.scatter(data_frame=pd.DataFrame(columns=["A"]), x="A", y="A"))],
        controls=[vm.Filter(id="test_filter", column="A")],
    )
    Vizro._pre_build()


# TODO: improve these tests to check that actual callback is registered.
# In general, we should aim to make all the tests in TestActionBuild and TestActionPrivateMethods
# higher level and more "real" so that we're testing actual callback execution and not just private
# helper methods.
class TestActionBuild:
    """Tests action build method."""

    def test_custom_action_build(self, identity_action_function, custom_action_build_expected):
        action = Action(id="action_test", function=identity_action_function())
        result = json.loads(json.dumps(action.build(), cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(custom_action_build_expected, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected

    @pytest.mark.usefixtures("managers_one_page_without_graphs_one_button")
    def test_predefined_export_data_action_build(self, predefined_action_build_expected):
        predefined_filter_action = model_manager["test_page"].controls[0].selector.actions[0].actions[0]
        result = json.loads(json.dumps(predefined_filter_action.build(), cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(predefined_action_build_expected, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected


class TestActionPrivateMethods:
    """Test action private methods."""

    def test_get_callback_mapping_no_inputs_no_outputs(self, identity_action_function):
        action = Action(function=identity_action_function())
        callback_inputs, callback_outputs, action_components = action._get_callback_mapping()
        assert callback_inputs == {}
        assert callback_outputs == {}
        assert action_components == []

    @pytest.mark.parametrize(
        "inputs_and_outputs, expected_get_callback_mapping_inputs, expected_get_callback_mapping_outputs",
        [
            (
                ["component.property"],
                [State("component", "property")],
                Output("component", "property"),
            ),
            (
                ["component_1.property", "component_2.property"],
                [State("component_1", "property"), State("component_2", "property")],
                [Output("component_1", "property"), Output("component_2", "property")],
            ),
        ],
    )
    def test_get_callback_mapping_with_inputs_and_outputs(  # pylint: disable=too-many-arguments
        self,
        inputs_and_outputs,
        identity_action_function,
        expected_get_callback_mapping_inputs,
        expected_get_callback_mapping_outputs,
    ):
        action = Action(function=identity_action_function(), inputs=inputs_and_outputs, outputs=inputs_and_outputs)
        callback_inputs, callback_outputs, action_components = action._get_callback_mapping()
        assert callback_inputs == expected_get_callback_mapping_inputs
        assert callback_outputs == expected_get_callback_mapping_outputs
        assert action_components == []

    @pytest.mark.parametrize("inputs", [["value"], {"arg": "value"}])
    def test_action_callback_function_inputs_args_or_kwargs(self, identity_action_function, inputs):
        action = Action(function=identity_action_function())
        assert action._action_callback_function(inputs=inputs, outputs=Output("component", "property")) == "value"

    @pytest.mark.parametrize(
        "custom_action_function_mock_return, callback_outputs",
        [
            # no outputs
            (None, []),
            (None, {}),
            (None, None),
            # single output - note this is not [Output(...)] in a list
            (None, Output("component", "property")),
            (["value_1", "value_2"], Output("component", "property")),
            ({"key_1": "value_1"}, Output("component", "property")),
            ("abc", Output("component", "property")),
            # multiple list outputs
            ("ab", [Output("component_1", "property"), Output("component_2", "property")]),
            (["value_1", "value_2"], [Output("component_1", "property"), Output("component_2", "property")]),
            (
                {"component_1": "value_1", "component_2_property": "value_2"},
                [Output("component", "property"), Output("component_2", "property")],
            ),
            # multiple dict outputs
            (
                {"component_1": "value_1"},
                {"component_1": Output("component", "property")},
            ),
            (
                {"component_1": "value_1", "component_2": "value_2"},
                {"component_1": Output("component_1", "property"), "component_2": Output("component_2", "property")},
            ),
        ],
        indirect=["custom_action_function_mock_return"],
    )
    def test_action_callback_function_return_value_valid(self, custom_action_function_mock_return, callback_outputs):
        action = Action(function=custom_action_function_mock_return())
        # If no error is raised by _action_callback_function then running it should return exactly the same
        # as the output of the custom_action_function_mock_return.
        assert (
            action._action_callback_function(inputs={}, outputs=callback_outputs)
            == custom_action_function_mock_return()()
        )

    @pytest.mark.parametrize("callback_outputs", [[], {}, None])
    @pytest.mark.parametrize(
        "custom_action_function_mock_return",
        [False, 0, "", [], (), {}],
        indirect=True,
    )
    def test_action_callback_function_no_outputs_return_value_not_None(
        self, custom_action_function_mock_return, callback_outputs
    ):
        action = Action(function=custom_action_function_mock_return())
        with pytest.raises(
            ValueError,
            match="Action function has returned a value but the action has no defined outputs.",
        ):
            action._action_callback_function(inputs={}, outputs=callback_outputs)

    @pytest.mark.parametrize(
        "custom_action_function_mock_return",
        [None, False, 0, 123],
        indirect=["custom_action_function_mock_return"],
    )
    def test_action_callback_function_outputs_list_return_value_not_collection(
        self, custom_action_function_mock_return
    ):
        # Note it's not possible for _action_callback_function to be called with a single Output in a list, like
        # [Output(...)]. This would always be done as Output(...) outside a list instead.
        action = Action(function=custom_action_function_mock_return())
        with pytest.raises(
            ValueError,
            match="Action function has not returned a list-like object but the action's defined outputs are a list.",
        ):
            action._action_callback_function(
                inputs={}, outputs=[Output("component_1", "property"), Output("component_2", "property")]
            )

    @pytest.mark.parametrize(
        "custom_action_function_mock_return",
        [None, False, 0, 123, "", [], ()],
        indirect=["custom_action_function_mock_return"],
    )
    def test_action_callback_function_outputs_mapping_return_value_not_mapping(
        self, custom_action_function_mock_return
    ):
        action = Action(function=custom_action_function_mock_return())
        with pytest.raises(
            ValueError,
            match="Action function has not returned a dictionary-like object "
            "but the action's defined outputs are a dictionary.",
        ):
            action._action_callback_function(inputs={}, outputs={"output": Output("component", "property")})

    @pytest.mark.parametrize(
        "custom_action_function_mock_return",
        ["", [], (), {}, "abc", [1, 2, 3], (1, 2, 3), {"a": 1, "b": 2, "c": 3}],
        indirect=["custom_action_function_mock_return"],
    )
    def test_action_callback_function_outputs_list_return_value_length_not_match(
        self, custom_action_function_mock_return
    ):
        action = Action(function=custom_action_function_mock_return())
        with pytest.raises(
            ValueError,
            match="Number of action's returned elements .+ does not match the number of action's defined outputs 2.",
        ):
            action._action_callback_function(
                inputs={}, outputs=[Output("component_1", "property"), Output("component_2", "property")]
            )

    @pytest.mark.parametrize(
        "custom_action_function_mock_return",
        [{}, {"another_output": 1}, {"output": 1, "another_output": 2}],
        indirect=["custom_action_function_mock_return"],
    )
    def test_action_callback_function_outputs_mapping_return_value_keys_not_match(
        self, custom_action_function_mock_return
    ):
        action = Action(function=custom_action_function_mock_return())
        with pytest.raises(
            ValueError,
            match="Keys of action's returned value .+ do not match the action's defined outputs {'output'}.",
        ):
            action._action_callback_function(inputs={}, outputs={"output": Output("component", "property")})



"""Unit tests for vizro.models.NavLink."""
import re

import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
import pytest
from asserts import STRIP_ALL, assert_component_equal
from dash import html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm

pytestmark = pytest.mark.usefixtures("prebuilt_two_page_dashboard")


class TestNavLinkInstantiation:
    """Tests NavLink model instantiation."""

    def test_nav_link_mandatory_only(self):
        nav_link = vm.NavLink(label="Label")

        assert hasattr(nav_link, "id")
        assert nav_link.label == "Label"
        assert nav_link.icon == ""
        assert nav_link.pages == []

    def test_nav_link_mandatory_and_optional(self, pages_as_list):
        nav_link = vm.NavLink(id="nav_link", icon="home", label="Homepage", pages=pages_as_list)

        assert nav_link.id == "nav_link"
        assert nav_link.label == "Homepage"
        assert nav_link.icon == "home"
        assert nav_link.pages == pages_as_list

    @pytest.mark.parametrize("icon", ["Bar Chart", "bar chart", "bar_chart", "Bar_Chart", " bar_chart "])
    def test_validate_icon(self, icon):
        nav_link = vm.NavLink(icon=icon, label="Label")
        assert nav_link.icon == "bar_chart"

    def test_nav_link_valid_pages_as_dict(self, pages_as_dict):
        nav_link = vm.NavLink(pages=pages_as_dict, label="Label")
        assert nav_link.pages == pages_as_dict

    def test_mandatory_label_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.NavLink()

    @pytest.mark.parametrize("pages", [{"Group": []}, []])
    def test_invalid_field_pages_no_ids_provided(self, pages):
        with pytest.raises(ValidationError, match="Ensure this value has at least 1 item."):
            vm.NavLink(pages=pages)

    def test_invalid_field_pages_wrong_input_type(self):
        with pytest.raises(ValidationError, match="str type expected"):
            vm.NavLink(pages=[vm.Page(title="Page 3", components=[vm.Button()])])

    @pytest.mark.parametrize("pages", [["non existent page"], {"Group": ["non existent page"]}])
    def test_invalid_page(self, pages):
        with pytest.raises(
            ValidationError, match=re.escape("Unknown page ID ['non existent page'] provided to argument 'pages'.")
        ):
            vm.NavLink(pages=pages)


class TestNavLinkPreBuildMethod:
    def test_nav_link(self, pages_as_dict):
        nav_link = vm.NavLink(label="Label", pages=pages_as_dict)
        nav_link.pre_build()
        assert isinstance(nav_link._nav_selector, vm.Accordion)
        assert nav_link._nav_selector.pages == pages_as_dict


@pytest.mark.parametrize("pages", ["pages_as_dict", "pages_as_list"])
class TestNavLinkBuildMethod:
    """Tests NavLink model build method."""

    common_args = {"offset": 4, "withArrow": True, "position": "bottom-start"}

    def test_nav_link_active(self, pages, request):
        pages = request.getfixturevalue(pages)
        nav_link = vm.NavLink(id="nav_link", label="Label", icon="icon", pages=pages)
        nav_link.pre_build()
        built_nav_link = nav_link.build(active_page_id="Page 1")
        expected_button = dbc.Button(
            children=[
                dmc.Tooltip(
                    label="Label",
                    children=[html.Span("icon", className="material-symbols-outlined")],
                    **self.common_args,
                )
            ],
            active=True,
            href="/",
            className="icon-button",
            id="nav_link",
        )
        assert_component_equal(built_nav_link["nav_link"], expected_button)
        assert_component_equal(built_nav_link["nav-panel"].children, [dbc.Accordion()], keys_to_strip=STRIP_ALL)

    def test_nav_link_not_active(self, pages, request):
        pages = request.getfixturevalue(pages)
        nav_link = vm.NavLink(id="nav_link", label="Label", icon="icon", pages=pages)
        nav_link.pre_build()
        built_nav_link = nav_link.build(active_page_id="Page 3")
        expected_button = dbc.Button(
            children=[
                dmc.Tooltip(
                    label="Label",
                    children=[html.Span("icon", className="material-symbols-outlined")],
                    **self.common_args,
                )
            ],
            active=False,
            href="/",
            className="icon-button",
            id="nav_link",
        )
        assert_component_equal(built_nav_link["nav_link"], expected_button)
        assert "nav-panel" not in built_nav_link



"""Pre-defined action function "_filter" to be reused in `action` parameter of VizroBaseModels."""

from typing import Any, Callable, Dict, List

import pandas as pd
from dash import ctx

from vizro.actions._actions_utils import (
    _get_modified_page_figures,
)
from vizro.managers._model_manager import ModelID
from vizro.models.types import capture


@capture("action")
def _filter(
    filter_column: str,
    targets: List[ModelID],
    filter_function: Callable[[pd.Series, Any], pd.Series],
    **inputs: Dict[str, Any],
) -> Dict[str, Any]:
    """Filters targeted charts/components on page by interaction with `Filter` control.

    Args:
        filter_column: Data column to filter
        targets: List of target component ids to apply filters on
        filter_function: Filter function to apply
        inputs: Dict mapping action function names with their inputs e.g.
            inputs = {'filters': [], 'parameters': ['gdpPercap'], 'filter_interaction': [], 'theme_selector': True}

    Returns:
        Dict mapping target component ids to modified charts/components e.g. {'my_scatter': Figure({})}
    """
    return _get_modified_page_figures(
        targets=targets,
        ctds_filter=ctx.args_grouping["external"]["filters"],
        ctds_filter_interaction=ctx.args_grouping["external"]["filter_interaction"],
        ctds_parameters=ctx.args_grouping["external"]["parameters"],
    )



import logging
from pathlib import Path
from typing import List

import dash
import flask

from vizro._constants import STATIC_URL_PREFIX
from vizro.managers import data_manager, model_manager
from vizro.models import Dashboard

logger = logging.getLogger(__name__)


class Vizro:
    """The main class of the `vizro` package."""

    def __init__(self, **kwargs):
        """Initializes Dash app, stored in `self.dash`.

        Args:
            kwargs: Passed through to `Dash.__init__`, e.g. `assets_folder`, `url_base_pathname`. See
                [Dash documentation](https://dash.plotly.com/reference#dash.dash) for possible arguments.
        """
        self.dash = dash.Dash(**kwargs, use_pages=True, pages_folder="", title="Vizro")
        self.dash.config.external_stylesheets.append(
            "https://fonts.googleapis.com/css2?family=Material+Symbols+Outlined"
        )

        # Include Vizro assets (in the static folder) as external scripts and stylesheets. We extend self.dash.config
        # objects so the user can specify additional external_scripts and external_stylesheets via kwargs.
        vizro_assets_folder = Path(__file__).with_name("static")
        static_url_path = self.dash.config.requests_pathname_prefix + STATIC_URL_PREFIX
        vizro_css = self._get_external_assets(static_url_path, vizro_assets_folder, "css")
        vizro_js = [
            {"src": path, "type": "module"}
            for path in self._get_external_assets(static_url_path, vizro_assets_folder, "js")
        ]
        self.dash.config.external_stylesheets.extend(vizro_css)
        self.dash.config.external_scripts.extend(vizro_js)

        # Serve all assets (including files other than css and js) that live in vizro_assets_folder at the
        # route /vizro. Based on code in Dash.init_app that serves assets_folder. This respects the case that the
        # dashboard is not hosted at the root of the server, e.g. http://www.example.com/dashboard/vizro.
        blueprint_prefix = self.dash.config.routes_pathname_prefix.replace("/", "_").replace(".", "_")
        self.dash.server.register_blueprint(
            flask.Blueprint(
                f"{blueprint_prefix}vizro_assets",
                self.dash.config.name,
                static_folder=vizro_assets_folder,
                static_url_path=static_url_path,
            )
        )

    def build(self, dashboard: Dashboard):
        """Builds the dashboard.

        Args:
            dashboard (Dashboard): [`Dashboard`][vizro.models.Dashboard] object.

        Returns:
            Vizro: App object
        """
        # Note Dash.index uses self.dash.title instead of self.dash.app.config.title.
        if dashboard.title:
            self.dash.title = dashboard.title

        # Note that model instantiation and pre_build are independent of Dash.
        self._pre_build()

        self.dash.layout = dashboard.build()

        return self

    def run(self, *args, **kwargs):  # if type annotated, mkdocstring stops seeing the class
        """Runs the dashboard.

        Args:
            args: Passed through to `dash.run`.
            kwargs: Passed through to `dash.run`.
        """
        data_manager._frozen_state = True
        model_manager._frozen_state = True

        self.dash.run(*args, **kwargs)

    @staticmethod
    def _pre_build():
        """Runs pre_build method on all models in the model_manager."""
        # Note that a pre_build method can itself add a model (e.g. an Action) to the model manager, and so we need to
        # iterate through set(model_manager) rather than model_manager itself or we loop through something that
        # changes size.
        # Any models that are created during the pre-build process *will not* themselves have pre_build run on them.
        # In future may add a second pre_build loop after the first one.
        for model_id in set(model_manager):
            model = model_manager[model_id]
            if hasattr(model, "pre_build"):
                model.pre_build()

    @staticmethod
    def _reset():
        """Private method that clears all state in the vizro app."""
        data_manager._clear()
        model_manager._clear()
        dash._callback.GLOBAL_CALLBACK_LIST = []
        dash._callback.GLOBAL_CALLBACK_MAP = {}
        dash._callback.GLOBAL_INLINE_SCRIPTS = []
        dash.page_registry.clear()
        dash._pages.CONFIG.clear()
        dash._pages.CONFIG.__dict__.clear()

    @staticmethod
    def _get_external_assets(new_path: str, folder: Path, extension: str) -> List[str]:
        """Returns a list of paths to assets with given extension in folder, prefixed with new_path.

        e.g. with new_path="/vizro", extension="css", folder="/path/to/vizro/vizro-core/src/vizro/static",
        we will get ["/vizro/css/accordion.css", "/vizro/css/button.css", ...].
        """
        return sorted((new_path / path.relative_to(folder)).as_posix() for path in folder.rglob(f"*.{extension}"))



import importlib.util
import logging
from collections.abc import Collection, Mapping
from pprint import pformat
from typing import Any, Dict, List, Union

from dash import Input, Output, State, callback, html

try:
    from pydantic.v1 import Field, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, validator

import vizro.actions
from vizro.managers._model_manager import ModelID
from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models.types import CapturedCallable

logger = logging.getLogger(__name__)


class Action(VizroBaseModel):
    """Action to be inserted into `actions` of relevant component.

    Args:
        function (CapturedCallable): See [`CapturedCallable`][vizro.models.types.CapturedCallable].
        inputs (List[str]): Inputs in the form `<component_id>.<property>` passed to the action function.
            Defaults to `[]`.
        outputs (List[str]): Outputs in the form `<component_id>.<property>` changed by the action function.
            Defaults to `[]`.
    """

    function: CapturedCallable = Field(..., import_path=vizro.actions)
    inputs: List[str] = Field(
        [],
        description="Inputs in the form `<component_id>.<property>` passed to the action function.",
        regex="^[^.]+[.][^.]+$",
    )
    outputs: List[str] = Field(
        [],
        description="Outputs in the form `<component_id>.<property>` changed by the action function.",
        regex="^[^.]+[.][^.]+$",
    )

    # TODO: Problem: generic Action model shouldn't depend on details of particular actions like export_data.
    # Possible solutions: make a generic mapping of action functions to validation functions or the imports they
    # require, and make the code here look up the appropriate validation using the function as key
    # This could then also involve other validations currently only carried out at run-time in pre-defined actions, such
    # as e.g. checking if the correct arguments have been provided to the file_format in export_data.
    @validator("function")
    def validate_predefined_actions(cls, function):
        if function._function.__name__ == "export_data":
            file_format = function._arguments.get("file_format")
            if file_format not in [None, "csv", "xlsx"]:
                raise ValueError(f'Unknown "file_format": {file_format}.' f' Known file formats: "csv", "xlsx".')
            if file_format == "xlsx":
                if importlib.util.find_spec("openpyxl") is None and importlib.util.find_spec("xlsxwriter") is None:
                    raise ModuleNotFoundError(
                        "You must install either openpyxl or xlsxwriter to export to xlsx format."
                    )
        return function

    def _get_callback_mapping(self):
        """Builds callback inputs and outputs for the Action model callback, and returns action required components.

        callback_inputs, and callback_outputs are "dash.State" and "dash.Output" objects made of three parts:
            1. User configured inputs/outputs - for custom actions,
            2. Vizro configured inputs/outputs - for predefined actions,
            3. Hardcoded inputs/outputs - for custom and predefined actions
                (enable callbacks to live inside the Action loop).

        Returns: List of required components (e.g. dcc.Download) for the Action model added to the `Dashboard`
            container. Those components represent the return value of the Action build method.
        """
        from vizro.actions._callback_mapping._get_action_callback_mapping import _get_action_callback_mapping

        callback_inputs: Union[List[State], Dict[str, State]]
        if self.inputs:
            callback_inputs = [State(*input.split(".")) for input in self.inputs]
        else:
            callback_inputs = _get_action_callback_mapping(action_id=ModelID(str(self.id)), argument="inputs")

        callback_outputs: Union[List[Output], Dict[str, Output]]
        if self.outputs:
            callback_outputs = [Output(*output.split("."), allow_duplicate=True) for output in self.outputs]

            # Need to use a single Output in the @callback decorator rather than a single element list for the case
            # of a single output. This means the action function can return a single value (e.g. "text") rather than a
            # single element list (e.g. ["text"]).
            if len(callback_outputs) == 1:
                callback_outputs = callback_outputs[0]
        else:
            callback_outputs = _get_action_callback_mapping(action_id=ModelID(str(self.id)), argument="outputs")

        action_components = _get_action_callback_mapping(action_id=ModelID(str(self.id)), argument="components")

        return callback_inputs, callback_outputs, action_components

    def _action_callback_function(
        self,
        inputs: Union[Dict[str, Any], List[Any]],
        outputs: Union[Dict[str, Output], List[Output], Output, None],
    ) -> Any:
        logger.debug(
            "===== Running action with id %s, function %s =====",
            self.id,
            self.function._function.__name__,
        )
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Action inputs:\n%s", pformat(inputs, depth=2, width=200))
            logger.debug("Action outputs:\n%s", pformat(outputs, width=200))

        if isinstance(inputs, Mapping):
            return_value = self.function(**inputs)
        else:
            return_value = self.function(*inputs)

        # Delegate all handling of the return_value and mapping to appropriate outputs to Dash - we don't modify
        # return_value to reshape it in any way. All we do is do some error checking to raise clearer error messages.
        if not outputs:
            if return_value is not None:
                raise ValueError("Action function has returned a value but the action has no defined outputs.")
        elif isinstance(outputs, dict):
            if not isinstance(return_value, Mapping):
                raise ValueError(
                    "Action function has not returned a dictionary-like object "
                    "but the action's defined outputs are a dictionary."
                )
            if set(outputs) != set(return_value):
                raise ValueError(
                    f"Keys of action's returned value {set(return_value) or {}} "
                    f"do not match the action's defined outputs {set(outputs) or {}})."
                )
        elif isinstance(outputs, list):
            if not isinstance(return_value, Collection):
                raise ValueError(
                    "Action function has not returned a list-like object but the action's defined outputs are a list."
                )
            if len(return_value) != len(outputs):
                raise ValueError(
                    f"Number of action's returned elements {len(return_value)} does not match the number"
                    f" of action's defined outputs {len(outputs)}."
                )

        # If no error has been raised then the return_value is good and is returned as it is.
        # This could be a list of outputs, dictionary of outputs or any single value including None.
        return return_value

    @_log_call
    def build(self):
        """Builds a callback for the Action model and returns required components for the callback.

        Returns:
            List of required components (e.g. dcc.Download) for the Action model added to the `Dashboard` container.
        """
        external_callback_inputs, external_callback_outputs, action_components = self._get_callback_mapping()
        callback_inputs = {
            "external": external_callback_inputs,
            "internal": {"trigger": Input({"type": "action_trigger", "action_name": self.id}, "data")},
        }
        callback_outputs = {
            "internal": {"action_finished": Output("action_finished", "data", allow_duplicate=True)},
        }

        # If there are no outputs then we don't want the external part of callback_outputs to exist at all.
        # This allows the action function to return None and match correctly on to the callback_outputs dictionary
        # The (probably better) alternative to this would be just to define a dummy output for all such functions
        # so that the external key always exists.
        # Note that it's still possible to explicitly return None as a value when an output is specified.
        if external_callback_outputs:
            callback_outputs["external"] = external_callback_outputs

        logger.debug(
            "===== Building callback for Action with id %s, function %s =====",
            self.id,
            self.function._function.__name__,
        )
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Callback inputs:\n%s", pformat(callback_inputs["external"], width=200))
            logger.debug("Callback outputs:\n%s", pformat(callback_outputs.get("external"), width=200))

        @callback(output=callback_outputs, inputs=callback_inputs, prevent_initial_call=True)
        def callback_wrapper(external: Union[List[Any], Dict[str, Any]], internal: Dict[str, Any]) -> Dict[str, Any]:
            return_value = self._action_callback_function(inputs=external, outputs=callback_outputs.get("external"))
            if "external" in callback_outputs:
                return {"internal": {"action_finished": None}, "external": return_value}
            return {"internal": {"action_finished": None}}

        return html.Div(children=action_components, id=f"{self.id}_action_model_components_div", hidden=True)



import pytest

from vizro.actions._actions_utils import _create_target_arg_mapping, _update_nested_graph_properties


class TestUpdateNestedGraphProperties:
    def test_update_nested_graph_properties_single_level(self):
        graph = {"color": "blue"}
        result = _update_nested_graph_properties(graph, "color", "red")
        expected = {"color": "red"}
        assert result == expected

    @pytest.mark.parametrize(
        "graph, dot_separated_strings, value, expected",
        [
            ({"node": {"label": "A", "color": "blue"}}, "node.color", "red", {"node": {"label": "A", "color": "red"}}),
            (
                {"nodes": {"A": {"color": "blue"}, "B": {"color": "green"}}},
                "nodes.A.color",
                "red",
                {"nodes": {"A": {"color": "red"}, "B": {"color": "green"}}},
            ),
        ],
    )
    def test_update_nested_graph_properties_multiple_levels(self, graph, dot_separated_strings, value, expected):
        result = _update_nested_graph_properties(graph, dot_separated_strings, value)
        assert result == expected

    @pytest.mark.parametrize(
        "graph, dot_separated_strings, value, expected",
        [
            (
                {"nodes": {"A": {"color": "blue"}, "B": {"color": "green"}}},
                "nodes.C.color",
                "red",
                {"nodes": {"A": {"color": "blue"}, "B": {"color": "green"}, "C": {"color": "red"}}},
            ),
            (
                {"nodes": {"A": {"color": "blue"}, "B": {"color": "red"}}},
                "nodes.B.value",
                "red",
                {"nodes": {"A": {"color": "blue"}, "B": {"color": "red", "value": "red"}}},
            ),
            (
                {"nodes": {"A": {"color": "blue"}, "B": {"color": "green"}}},
                "nodes.B",
                "red",
                {"nodes": {"A": {"color": "blue"}, "B": "red"}},
            ),
            ({}, "color", "red", {"color": "red"}),
        ],
    )
    def test_update_nested_graph_properties_add_or_overwrite_keys(self, graph, dot_separated_strings, value, expected):
        result = _update_nested_graph_properties(graph, dot_separated_strings, value)
        assert result == expected

    def test_update_nested_graph_properties_invalid_type(self):
        graph = {"color": "blue"}
        with pytest.raises(TypeError, match="'str' object does not support item assignment"):
            _update_nested_graph_properties(graph, "color.value", "42")


class TestCreateTargetArgMapping:
    def test_single_string_one_component(self):
        input_strings = ["component1.argument1"]
        expected = {"component1": ["argument1"]}
        result = _create_target_arg_mapping(input_strings)
        assert result == expected

    def test_multiple_strings_different_components(self):
        input_strings = ["component1.argument1", "component2.argument2", "component1.argument3"]
        expected = {"component1": ["argument1", "argument3"], "component2": ["argument2"]}
        result = _create_target_arg_mapping(input_strings)
        assert result == expected

    def test_multiple_strings_same_component(self):
        input_strings = ["component1.argument1", "component1.argument2", "component1.argument3"]
        expected = {"component1": ["argument1", "argument2", "argument3"]}
        result = _create_target_arg_mapping(input_strings)
        assert result == expected

    def test_empty_input_list(self):
        input_strings = []
        expected = {}
        result = _create_target_arg_mapping(input_strings)
        assert result == expected

    def test_strings_without_separator(self):
        input_strings = ["component1_argument1", "component2_argument2"]
        with pytest.raises(ValueError, match="must contain a '.'"):
            _create_target_arg_mapping(input_strings)

    def test_strings_with_multiple_separators(self):
        input_strings = ["component1.argument1.extra", "component2.argument2.extra"]
        expected = {"component1": ["argument1.extra"], "component2": ["argument2.extra"]}
        assert _create_target_arg_mapping(input_strings) == expected



"""Unit tests for vizro.models.Navigation."""
import re

import dash_bootstrap_components as dbc
import pytest
from asserts import STRIP_ALL, assert_component_equal
from dash import html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm

pytestmark = pytest.mark.usefixtures("prebuilt_two_page_dashboard")


class TestNavigationInstantiation:
    def test_navigation_mandatory_only(self):
        navigation = vm.Navigation()

        assert hasattr(navigation, "id")
        assert navigation.pages == []
        assert navigation.nav_selector is None

    def test_navigation_mandatory_and_optional(self):
        accordion = vm.Accordion()
        navigation = vm.Navigation(id="navigation", pages=["Page 1", "Page 2"], nav_selector=accordion)

        assert navigation.id == "navigation"
        assert navigation.pages == ["Page 1", "Page 2"]
        assert navigation.nav_selector == accordion

    @pytest.mark.parametrize("pages", [{"Group": []}, []])
    def test_invalid_field_pages_no_ids_provided(self, pages):
        with pytest.raises(ValidationError, match="Ensure this value has at least 1 item."):
            vm.Navigation(pages=pages)

    def test_invalid_field_pages_wrong_input_type(self):
        with pytest.raises(ValidationError, match="str type expected"):
            vm.Navigation(pages=[vm.Page(title="Page 3", components=[vm.Button()])])

    @pytest.mark.parametrize("pages", [["non existent page"], {"Group": ["non existent page"]}])
    def test_invalid_page(self, pages):
        with pytest.raises(
            ValidationError, match=re.escape("Unknown page ID ['non existent page'] provided to argument 'pages'.")
        ):
            vm.Navigation(pages=pages)


class TestNavigationPreBuildMethod:
    def test_default_nav_selector(self, pages_as_dict):
        navigation = vm.Navigation(pages=pages_as_dict)
        navigation.pre_build()
        assert isinstance(navigation.nav_selector, vm.Accordion)
        assert navigation.nav_selector.pages == pages_as_dict

    def test_default_nav_selector_with_pages(self, pages_as_dict):
        navigation = vm.Navigation(pages=pages_as_dict, nav_selector=vm.Accordion(pages={"Group": ["Page 1"]}))
        navigation.pre_build()
        assert isinstance(navigation.nav_selector, vm.Accordion)
        assert navigation.nav_selector.pages == {"Group": ["Page 1"]}

    def test_non_default_nav_selector(self, pages_as_dict):
        navigation = vm.Navigation(pages=pages_as_dict, nav_selector=vm.NavBar())
        navigation.pre_build()
        assert isinstance(navigation.nav_selector, vm.NavBar)
        assert navigation.nav_selector.pages == pages_as_dict

    def test_non_default_nav_selector_with_pages(self, pages_as_dict):
        navigation = vm.Navigation(pages=pages_as_dict, nav_selector=vm.NavBar(pages={"Group": ["Page 1"]}))
        navigation.pre_build()
        assert isinstance(navigation.nav_selector, vm.NavBar)
        assert navigation.nav_selector.pages == {"Group": ["Page 1"]}


class TestNavigationBuildMethod:
    """Tests navigation model build method."""

    @pytest.mark.parametrize("pages", ["pages_as_dict", "pages_as_list"])
    def test_default_nav_selector(self, pages, request):
        pages = request.getfixturevalue(pages)
        navigation = vm.Navigation(pages=pages)
        navigation.pre_build()
        built_navigation = navigation.build(active_page_id="Page 1")
        assert_component_equal(built_navigation["nav-bar"], html.Div(hidden=True, id="nav-bar"))
        assert_component_equal(built_navigation["nav-panel"], html.Div(id="nav-panel"), keys_to_strip={"children"})
        assert_component_equal(built_navigation["nav-panel"].children, [dbc.Accordion()], keys_to_strip=STRIP_ALL)

    def test_non_default_nav_selector_pags_as_dict(self, pages_as_dict, built_nav_link=None):
        navigation = vm.Navigation(pages=pages_as_dict, nav_selector=vm.NavBar())
        navigation.pre_build()
        built_navigation = navigation.build(active_page_id="Page 1")
        assert_component_equal(built_navigation["nav-bar"], html.Div(id="nav-bar"), keys_to_strip={"children"})
        assert_component_equal(built_navigation["nav-panel"], html.Div(id="nav-panel"), keys_to_strip={"children"})
        assert_component_equal(built_navigation["nav-panel"].children, [dbc.Accordion()], keys_to_strip=STRIP_ALL)

    def test_non_default_nav_selector_pages_as_list(self, pages_as_list):
        navigation = vm.Navigation(pages=pages_as_list, nav_selector=vm.NavBar())
        navigation.pre_build()
        built_navigation = navigation.build(active_page_id="Page 1")
        assert_component_equal(built_navigation["nav-bar"], html.Div(id="nav-bar"), keys_to_strip={"children"})
        assert_component_equal(
            built_navigation["nav-panel"], html.Div(id="nav-panel", hidden=True), keys_to_strip={"children"}
        )



from __future__ import annotations

from collections.abc import Mapping
from typing import Dict, List, Literal

from dash import html

try:
    from pydantic.v1 import Field, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, validator

from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models._navigation._navigation_utils import _NavBuildType, _validate_pages
from vizro.models._navigation.nav_link import NavLink


class NavBar(VizroBaseModel):
    """Navigation bar to be used as a nav_selector for `Navigation`.

    Args:
        type (Literal["nav_bar"]): Defaults to `"nav_bar"`.
        pages (Dict[str, List[str]]): Mapping from name of a pages group to a list of page IDs. Defaults to `{}`.
        items (List[NavLink]): See [`NavLink`][vizro.models.NavLink]. Defaults to `[]`.
    """

    type: Literal["nav_bar"] = "nav_bar"
    pages: Dict[str, List[str]] = Field({}, description="Mapping from name of a pages group to a list of page IDs.")
    items: List[NavLink] = []

    # validators
    _validate_pages = validator("pages", allow_reuse=True)(_validate_pages)

    @validator("pages", pre=True)
    def coerce_pages_type(cls, pages):
        if isinstance(pages, Mapping):
            return pages
        return {page: [page] for page in pages}

    @_log_call
    def pre_build(self):
        self.items = self.items or [
            NavLink(label=group_title, pages=pages) for group_title, pages in self.pages.items()
        ]

        for position, item in enumerate(self.items, 1):
            # The filter icons are named filter_1, filter_2, etc. up to filter_9. If there are more than 9 items, then
            # the 10th and all subsequent items are named filter_9+.
            item.icon = item.icon or f"filter_{position}" if position <= 9 else "filter_9+"  # noqa: PLR2004

        # Since models instantiated in pre_build do not themselves have pre_build called on them, we call it manually
        # here.
        for item in self.items:
            item.pre_build()

        return self.items

    @_log_call
    def build(self, *, active_page_id=None) -> _NavBuildType:
        # We always show all the nav_link buttons, but only show the accordion for the active page. This works because
        # item.build only returns the nav_panel Div when the item is active.
        # In future maybe we should do this by showing all navigation panels and then setting hidden=True for all but
        # one using a clientside callback?
        # Wrapping built_items into html.Div here is not for rendering purposes but so that we can look up the
        # components by id easily instead of needing to iterate through a nested list.
        built_items = html.Div([item.build(active_page_id=active_page_id) for item in self.items])
        buttons = [built_items[item.id] for item in self.items]
        if "nav-panel" in built_items:
            nav_panel = built_items["nav-panel"]
        else:
            # Active page is not in navigation at all, so hide navigation panel.
            nav_panel = html.Div(hidden=True, id="nav-panel")

        return html.Div([html.Div(buttons, id="nav-bar"), nav_panel])



"""Fixtures to be shared across several tests."""

import pytest

import vizro.models as vm


@pytest.fixture()
def pages_as_list():
    return ["Page 1", "Page 2"]


@pytest.fixture
def pages_as_dict():
    return {"Group": ["Page 1", "Page 2"]}


@pytest.fixture()
def prebuilt_two_page_dashboard(vizro_app, page_1, page_2):
    dashboard = vm.Dashboard(pages=[page_1, page_2])
    dashboard.pre_build()
    return dashboard



from __future__ import annotations

import logging
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, List, Literal, TypedDict

import dash
import dash_bootstrap_components as dbc
import dash_daq as daq
from dash import ClientsideFunction, Input, Output, clientside_callback, get_relative_path, html

try:
    from pydantic.v1 import Field, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, validator

from dash.development.base_component import Component

import vizro
from vizro._constants import MODULE_PAGE_404, STATIC_URL_PREFIX
from vizro.actions._action_loop._action_loop import ActionLoop
from vizro.models import Navigation, VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models._navigation._navigation_utils import _NavBuildType

if TYPE_CHECKING:
    from vizro.models import Page
    from vizro.models._page import _PageBuildType

logger = logging.getLogger(__name__)


def _all_hidden(components: List[Component]):
    """Returns True if all components are either None and/or have hidden=True."""
    return all(component is None or getattr(component, "hidden", False) for component in components)


# This is just used for type checking. Ideally it would inherit from some dash.development.base_component.Component
# (e.g. html.Div) as well as TypedDict, but that's not possible, and Dash does not have typing support anyway. When
# this type is used, the object is actually still a dash.development.base_component.Component, but this makes it easier
# to see what contract the component fulfills by making the expected keys explicit.
_PageDivsType = TypedDict(
    "_PageDivsType",
    {
        "dashboard-title": html.Div,
        "settings": html.Div,
        "page-title": html.H2,
        "nav-bar": html.Div,
        "nav-panel": html.Div,
        "control-panel": html.Div,
        "components": html.Div,
    },
)


class Dashboard(VizroBaseModel):
    """Vizro Dashboard to be used within [`Vizro`][vizro._vizro.Vizro.build].

    Args:
        pages (List[Page]): See [`Page`][vizro.models.Page].
        theme (Literal["vizro_dark", "vizro_light"]): Layout theme to be applied across dashboard.
            Defaults to `vizro_dark`.
        navigation (Navigation): See [`Navigation`][vizro.models.Navigation]. Defaults to `None`.
        title (str): Dashboard title to appear on every page on top left-side. Defaults to `""`.
    """

    pages: List[Page]
    theme: Literal["vizro_dark", "vizro_light"] = Field(
        "vizro_dark", description="Layout theme to be applied across dashboard. Defaults to `vizro_dark`"
    )
    navigation: Navigation = None  # type: ignore[assignment]
    title: str = Field("", description="Dashboard title to appear on every page on top left-side.")

    @validator("pages", always=True)
    def validate_pages(cls, pages):
        if not pages:
            raise ValueError("Ensure this value has at least 1 item.")
        return pages

    @validator("navigation", always=True)
    def set_navigation_pages(cls, navigation, values):
        if "pages" not in values:
            return navigation

        navigation = navigation or Navigation()
        navigation.pages = navigation.pages or [page.id for page in values["pages"]]
        return navigation

    @_log_call
    def pre_build(self):
        meta_image = self._infer_image("app") or self._infer_image("logo")

        # Setting order here ensures that the pages in dash.page_registry preserves the order of the List[Page].
        # For now the homepage (path /) corresponds to self.pages[0].
        # Note redirect_from=["/"] doesn't work and so the / route must be defined separately.
        for order, page in enumerate(self.pages):
            dash.register_page(
                module=page.id,
                name=page.title,
                description=page.description,
                image=meta_image,
                title=f"{self.title}: {page.title}" if self.title else page.title,
                path=page.path if order else "/",
                order=order,
                layout=partial(self._make_page_layout, page),
            )
        dash.register_page(module=MODULE_PAGE_404, layout=self._make_page_404_layout())

    @_log_call
    def build(self):
        for page in self.pages:
            page.build()  # TODO: ideally remove, but necessary to register slider callbacks

        clientside_callback(
            ClientsideFunction(namespace="clientside", function_name="update_dashboard_theme"),
            Output("dashboard_container_outer", "className"),
            Input("theme_selector", "on"),
        )

        return html.Div(
            id="dashboard_container_outer",
            children=[
                html.Div(vizro.__version__, id="vizro_version", hidden=True),
                ActionLoop._create_app_callbacks(),
                dash.page_container,
            ],
            className=self.theme,
        )

    def _get_page_divs(self, page: Page) -> _PageDivsType:
        # Identical across pages
        dashboard_title = (
            html.H2(self.title, id="dashboard-title") if self.title else html.H2(hidden=True, id="dashboard-title")
        )

        settings = html.Div(
            daq.BooleanSwitch(
                id="theme_selector", on=self.theme == "vizro_dark", persistence=True, persistence_type="session"
            ),
            id="settings",
        )

        # Shared across pages but slightly differ in content. These could possibly be done by a clientside
        # callback instead.
        page_title = html.H2(page.title, id="page-title")
        navigation: _NavBuildType = self.navigation.build(active_page_id=page.id)
        nav_bar = navigation["nav-bar"]
        nav_panel = navigation["nav-panel"]

        # Different across pages
        page_content: _PageBuildType = page.build()
        control_panel = page_content["control-panel"]
        components = page_content["components"]
        return html.Div([dashboard_title, settings, page_title, nav_bar, nav_panel, control_panel, components])

    def _arrange_page_divs(self, page_divs: _PageDivsType):
        left_header_divs = [page_divs["dashboard-title"]]
        left_sidebar_divs = [page_divs["nav-bar"]]
        left_main_divs = [
            html.Div(left_header_divs, id="left-header", hidden=_all_hidden(left_header_divs)),
            page_divs["nav-panel"],
            page_divs["control-panel"],
        ]

        left_sidebar = html.Div(left_sidebar_divs, id="left-sidebar", hidden=_all_hidden(left_sidebar_divs))
        left_main = html.Div(left_main_divs, id="left-main", hidden=_all_hidden(left_main_divs))
        left_side = html.Div([left_sidebar, left_main], id="left-side")

        right_header = html.Div([page_divs["page-title"], page_divs["settings"]], id="right-header")
        right_main = page_divs["components"]
        right_side = html.Div([right_header, right_main], id="right-side")

        return html.Div([left_side, right_side], id="page-container")

    def _make_page_layout(self, page: Page):
        page_divs = self._get_page_divs(page=page)
        return self._arrange_page_divs(page_divs=page_divs)

    @staticmethod
    def _make_page_404_layout():
        return html.Div(
            [
                html.Img(src=get_relative_path(f"/{STATIC_URL_PREFIX}/images/errors/error_404.svg")),
                html.Div(
                    [
                        html.Div(
                            [
                                html.H3("This page could not be found.", className="heading-3-600"),
                                html.P("Make sure the URL you entered is correct."),
                            ],
                            className="error_text_container",
                        ),
                        dbc.Button("Take me home", href=get_relative_path("/"), className="button_primary"),
                    ],
                    className="error_content_container",
                ),
            ],
            className="page_error_container",
        )

    def _infer_image(self, filename: str):
        valid_extensions = [".apng", ".avif", ".gif", ".jpeg", ".jpg", ".png", ".svg", ".webp"]
        assets_folder = Path(dash.get_app().config.assets_folder)
        if assets_folder.is_dir():
            for path in Path(assets_folder).rglob(f"{filename}.*"):
                if path.suffix in valid_extensions:
                    return str(path.relative_to(assets_folder))



"""The data manager handles access to all DataFrames used in a Vizro app."""
from typing import Callable, Dict, Union

import pandas as pd

from vizro.managers._managers_utils import _state_modifier

# Really ComponentID and DatasetName should be NewType and not just aliases but then for a user's code to type check
# correctly they would need to cast all strings to these types.
ComponentID = str
DatasetName = str
pd_LazyDataFrame = Callable[[], pd.DataFrame]


class DataManager:
    """Object to handle all data for the `vizro` application.

    Examples:
        >>> import plotly.express as px
        >>> data_manager["iris"] = px.data.iris()

    """

    def __init__(self):
        self.__lazy_data: Dict[DatasetName, pd_LazyDataFrame] = {}
        self.__original_data: Dict[DatasetName, pd.DataFrame] = {}
        self.__component_to_original: Dict[ComponentID, DatasetName] = {}
        self._frozen_state = False

    @_state_modifier
    def __setitem__(self, dataset_name: DatasetName, data: Union[pd.DataFrame, pd_LazyDataFrame]):
        """Adds `data` to the `DataManager` with key `dataset_name`.

        This is the only user-facing function when configuring a simple dashboard. Others are only used internally
        in Vizro or advanced users who write their own actions.
        """
        if dataset_name in self.__original_data or dataset_name in self.__lazy_data:
            raise ValueError(f"Dataset {dataset_name} already exists.")

        if callable(data):
            self.__lazy_data[dataset_name] = data
        elif isinstance(data, pd.DataFrame):
            self.__original_data[dataset_name] = data
        else:
            raise TypeError(
                f"Dataset {dataset_name} must be a pandas DataFrame or callable that returns pandas DataFrame."
            )

    @_state_modifier
    def _add_component(self, component_id: ComponentID, dataset_name: DatasetName):
        """Adds a mapping from `component_id` to `dataset_name`."""
        if dataset_name not in self.__original_data and dataset_name not in self.__lazy_data:
            raise KeyError(f"Dataset {dataset_name} does not exist.")
        if component_id in self.__component_to_original:
            raise ValueError(
                f"Component with id={component_id} already exists and is mapped to dataset "
                f"{self.__component_to_original[component_id]}. Components must uniquely map to a dataset across the "
                f"whole dashboard. If you are working from a Jupyter Notebook, please either restart the kernel, or "
                f"use 'from vizro import Vizro; Vizro._reset()`."
            )
        self.__component_to_original[component_id] = dataset_name

    def _get_component_data(self, component_id: ComponentID) -> pd.DataFrame:
        """Returns the original data for `component_id`."""
        if component_id not in self.__component_to_original:
            raise KeyError(f"Component {component_id} does not exist. You need to call add_component first.")
        dataset_name = self.__component_to_original[component_id]

        # Populate original data on first access only
        if dataset_name not in self.__original_data:
            self.__original_data[dataset_name] = self.__lazy_data[dataset_name]()

        # Return a copy so that the original data cannot be modified. This is not necessary if we are careful
        # to not do any inplace=True operations, but probably safest to leave it here.
        return self.__original_data[dataset_name].copy()

    def _has_registered_data(self, component_id: ComponentID) -> bool:
        try:
            self._get_component_data(component_id)
            return True
        except KeyError:
            return False

    def _clear(self):
        self.__init__()  # type: ignore[misc]


data_manager = DataManager()


if __name__ == "__main__":
    from functools import partial

    import vizro.plotly.express as px

    dm = data_manager
    dm["iris"] = px.data.iris()

    dm._add_component("component_id_a", "iris")
    print(len(dm._get_component_data("component_id_a")))  # 150   # noqa: T201

    dm._add_component("component_id_b", "iris")
    df_a = dm._get_component_data("component_id_a")
    df_a.drop(columns="species", inplace=True)
    print(df_a.shape)  # (150, 5)   # noqa: T201
    df_b = dm._get_component_data("component_id_b")
    print(df_b.shape)  # (150, 6)   # noqa: T201

    # Lazy loading example 1
    def retrieve_iris():
        df = px.data.iris()
        subset = df.query("species == 'setosa'")
        return subset

    dm["iris_subset"] = retrieve_iris
    dm._add_component("component_id_c", "iris_subset")
    print(len(dm._get_component_data("component_id_c")))  # 50   # noqa: T201

    # Lazy loading example 2
    def retrieve_one_species(species):
        df = px.data.iris()
        subset = df[df["species"] == species].copy()
        return subset

    dm["data_from_external_1"] = lambda: retrieve_one_species("setosa")
    dm._add_component("component_id_d", "data_from_external_1")
    print(len(dm._get_component_data("component_id_d")))  # 50   # noqa: T201

    # Lazy loading example 3
    dm["data_from_external_2"] = partial(retrieve_one_species, "setosa")
    dm._add_component("component_id_e", "data_from_external_2")
    print(len(dm._get_component_data("component_id_e")))  # 50   # noqa: T201



from __future__ import annotations

from dash import html

try:
    from pydantic.v1 import validator
except ImportError:  # pragma: no cov
    from pydantic import validator

from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models._navigation._navigation_utils import _NavBuildType, _validate_pages
from vizro.models._navigation.accordion import Accordion
from vizro.models.types import NavPagesType, NavSelectorType


class Navigation(VizroBaseModel):
    """Navigation in [`Dashboard`][vizro.models.Dashboard] to structure [`Pages`][vizro.models.Page].

    Args:
        pages (NavPagesType): See [`NavPagesType`][vizro.models.types.NavPagesType]. Defaults to `[]`.
        nav_selector (NavSelectorType): See [`NavSelectorType`][vizro.models.types.NavSelectorType].
            Defaults to `None`.
    """

    pages: NavPagesType = []
    nav_selector: NavSelectorType = None

    # validators
    _validate_pages = validator("pages", allow_reuse=True)(_validate_pages)

    @_log_call
    def pre_build(self):
        # Since models instantiated in pre_build do not themselves have pre_build called on them, we call it manually
        # here. Note that not all nav_selectors have pre_build (Accordion does not).
        self.nav_selector = self.nav_selector or Accordion()
        self.nav_selector.pages = self.nav_selector.pages or self.pages
        if hasattr(self.nav_selector, "pre_build"):
            self.nav_selector.pre_build()

    @_log_call
    def build(self, *, active_page_id=None) -> _NavBuildType:
        nav_selector = self.nav_selector.build(active_page_id=active_page_id)

        if "nav-bar" not in nav_selector:
            # e.g. nav_selector is Accordion and nav_selector.build returns single html.Div with id="nav-panel".
            # This will make it match the case e.g. nav_selector is NavBar and nav_selector.build returns html.Div
            # containing children with id="nav-bar" and id="nav-panel"
            nav_selector = html.Div([html.Div(hidden=True, id="nav-bar"), nav_selector])

        return nav_selector



"""Pre-defined action function "export_data" to be reused in `action` parameter of VizroBaseModels."""

from typing import Any, Dict, List, Optional

from dash import ctx, dcc
from typing_extensions import Literal

from vizro.actions._actions_utils import (
    _get_filtered_data,
)
from vizro.managers import model_manager
from vizro.managers._model_manager import ModelID
from vizro.models.types import capture


@capture("action")
def export_data(
    targets: Optional[List[ModelID]] = None,
    file_format: Literal["csv", "xlsx"] = "csv",
    **inputs: Dict[str, Any],
) -> Dict[str, Any]:
    """Exports visible data of target charts/components on page after being triggered.

    Args:
        targets: List of target component ids to download data from. Defaults to `None`.
        file_format: Format of downloaded files. Defaults to `csv`.
        inputs: Dict mapping action function names with their inputs e.g.
            inputs = {'filters': [], 'parameters': ['gdpPercap'], 'filter_interaction': [], 'theme_selector': True}

    Raises:
        ValueError: If unknown file extension is provided.
        ValueError: If target component does not exist on page.

    Returns:
        Dict mapping target component id to modified charts/components e.g. {'my_scatter': Figure({})}
    """
    if not targets:
        targets = [
            output["id"]["target_id"]
            for output in ctx.outputs_list
            if isinstance(output["id"], dict) and output["id"]["type"] == "download_dataframe"
        ]
    for target in targets:
        if target not in model_manager:
            raise ValueError(f"Component '{target}' does not exist.")

    data_frames = _get_filtered_data(
        targets=targets,
        ctds_filters=ctx.args_grouping["external"]["filters"],
        ctds_filter_interaction=ctx.args_grouping["external"]["filter_interaction"],
    )

    outputs = {}
    for target_id in targets:
        if file_format == "csv":
            writer = data_frames[target_id].to_csv
        elif file_format == "xlsx":
            writer = data_frames[target_id].to_excel
        # Invalid file_format should be caught by Action validation

        outputs[f"download_dataframe_{target_id}"] = dcc.send_data_frame(
            writer=writer, filename=f"{target_id}.{file_format}", index=False
        )

    return outputs



import pytest

import vizro.models as vm
from vizro.managers import model_manager
from vizro.models._action._actions_chain import ActionsChain
from vizro.models._controls.parameter import Parameter
from vizro.models.types import CapturedCallable


@pytest.mark.usefixtures("managers_one_page_two_graphs")
class TestParameterInstantiation:
    def test_instantiation(self):
        parameter = Parameter(
            targets=["scatter_chart.x"],
            selector=vm.Dropdown(
                options=["lifeExp", "gdpPercap", "pop"], multi=False, value="lifeExp", title="Choose x-axis"
            ),
        )
        assert parameter.type == "parameter"
        assert parameter.targets == ["scatter_chart.x"]
        assert parameter.selector.type == "dropdown"

    def test_check_dot_notation_failed(self):
        with pytest.raises(
            ValueError,
            match="Invalid target scatter_chart. Targets must be supplied in the from of "
            "<target_component>.<target_argument>",
        ):
            Parameter(
                targets=["scatter_chart"],
                selector=vm.Dropdown(options=["lifeExp", "pop"]),
            )

    def test_check_target_present_failed(self):
        with pytest.raises(ValueError, match="Target scatter_chart_invalid not found in model_manager."):
            Parameter(
                targets=["scatter_chart_invalid.x"],
                selector=vm.Dropdown(options=["lifeExp", "pop"]),
            )

    def test_duplicate_parameter_target_failed(self):
        with pytest.raises(ValueError, match="Duplicate parameter targets {'scatter_chart.x'} found."):
            Parameter(
                targets=["scatter_chart.x", "scatter_chart.x"],
                selector=vm.Dropdown(options=["lifeExp", "pop"]),
            )

    def test_duplicate_parameter_target_failed_two_params(self):
        with pytest.raises(ValueError, match="Duplicate parameter targets {'scatter_chart.x'} found."):
            Parameter(
                targets=["scatter_chart.x"],
                selector=vm.Dropdown(options=["lifeExp", "pop"]),
            )
            Parameter(
                targets=["scatter_chart.x"],
                selector=vm.Dropdown(options=["lifeExp", "pop"]),
            )


@pytest.mark.usefixtures("managers_one_page_two_graphs")
class TestPreBuildMethod:
    @pytest.mark.parametrize(
        "test_input, title",
        [
            (vm.Checklist(options=["lifeExp", "gdpPercap", "pop"], value=["lifeExp"]), "x"),
            (vm.Dropdown(options=["lifeExp", "gdpPercap", "pop"], multi=False, value="lifeExp"), "x"),
            (
                vm.RadioItems(options=["lifeExp", "gdpPercap", "pop"], value="lifeExp", title="Choose x-axis"),
                "Choose x-axis",
            ),
        ],
    )
    def test_set_target_and_title_valid(self, test_input, title):
        parameter = Parameter(targets=["scatter_chart.x"], selector=test_input)
        page = model_manager["test_page"]
        page.controls = [parameter]
        parameter.pre_build()
        assert parameter.targets == ["scatter_chart.x"]
        assert parameter.selector.title == title

    @pytest.mark.parametrize(
        "test_input",
        [
            (vm.Checklist(options=["lifeExp", "gdpPercap", "pop"], value=["lifeExp"])),
            (vm.Dropdown(options=["lifeExp", "gdpPercap", "pop"], multi=False, value="lifeExp")),
            (vm.RadioItems(options=["lifeExp", "gdpPercap", "pop"], value="lifeExp")),
        ],
    )
    def test_set_actions(self, test_input):
        parameter = Parameter(targets=["scatter_chart.x"], selector=test_input)
        page = model_manager["test_page"]
        page.controls = [parameter]
        parameter.pre_build()
        default_action = parameter.selector.actions[0]
        assert isinstance(default_action, ActionsChain)
        assert isinstance(default_action.actions[0].function, CapturedCallable)
        assert default_action.actions[0].id == f"parameter_action_{parameter.id}"

    @pytest.mark.parametrize("test_input", [vm.Slider(), vm.RangeSlider()])
    def test_set_slider_values_invalid(self, test_input):
        parameter = Parameter(targets=["scatter_chart.x"], selector=test_input)
        page = model_manager["test_page"]
        page.controls = [parameter]
        with pytest.raises(
            TypeError, match=f"{test_input.type} requires the arguments 'min' and 'max' when used within Parameter."
        ):
            parameter.pre_build()

    @pytest.mark.parametrize("test_input", [vm.Checklist(), vm.Dropdown(), vm.RadioItems()])
    def test_set_categorical_selectors_with_missing_options(self, test_input):
        parameter = Parameter(targets=["scatter_chart.x"], selector=test_input)
        page = model_manager["test_page"]
        page.controls = [parameter]
        with pytest.raises(
            TypeError, match=f"{parameter.selector.type} requires the argument 'options' when used within Parameter."
        ):
            parameter.pre_build()


@pytest.mark.usefixtures("managers_one_page_two_graphs")
class TestParameterBuild:
    """Tests parameter build method."""

    @pytest.mark.parametrize(
        "test_input",
        [
            vm.Checklist(options=["lifeExp", "gdpPercap", "pop"]),
            vm.Dropdown(options=["lifeExp", "gdpPercap", "pop"]),
            vm.RadioItems(options=["lifeExp", "gdpPercap", "pop"]),
        ],
    )
    def test_build_parameter(self, test_input):
        parameter = Parameter(targets=["scatter_chart.x"], selector=test_input)
        page = model_manager["test_page"]
        page.controls = [parameter]
        parameter.pre_build()
        result = str(parameter.build())
        expected = str(test_input.build())
        assert result == expected



from typing import List, Literal, Optional, Union

from dash import dcc, html

try:
    from pydantic.v1 import Field, PrivateAttr, root_validator, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, root_validator, validator

from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._components.form._form_utils import get_options_and_default, validate_options_dict, validate_value
from vizro.models._models_utils import _log_call
from vizro.models.types import MultiValueType, OptionsType, SingleValueType


class Dropdown(VizroBaseModel):
    """Categorical multi-selector `Dropdown` to be provided to [`Filter`][vizro.models.Filter].

    Args:
        type (Literal["dropdown"]): Defaults to `"dropdown"`.
        options (OptionsType): See [`OptionsType`][vizro.models.types.OptionsType]. Defaults to `[]`.
        value (Optional[Union[SingleValueType, MultiValueType]]): See
            [`SingleValueType`][vizro.models.types.SingleValueType] and
            [`MultiValueType`][vizro.models.types.MultiValueType]. Defaults to `None`.
        multi (bool): Whether to allow selection of multiple values. Defaults to `True`.
        title (str): Title to be displayed. Defaults to `""`.
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.
    """

    type: Literal["dropdown"] = "dropdown"
    options: OptionsType = []
    value: Optional[Union[SingleValueType, MultiValueType]] = None
    multi: bool = Field(True, description="Whether to allow selection of multiple values")
    title: str = Field("", description="Title to be displayed")
    actions: List[Action] = []

    # Component properties for actions and interactions
    _input_property: str = PrivateAttr("value")

    # Re-used validators
    _set_actions = _action_validator_factory("value")
    _validate_options = root_validator(allow_reuse=True, pre=True)(validate_options_dict)
    _validate_value = validator("value", allow_reuse=True, always=True)(validate_value)

    @validator("multi", always=True)
    def validate_multi(cls, multi, values):
        if "value" not in values:
            return multi

        if values["value"] and multi is False and isinstance(values["value"], list):
            raise ValueError("Please set multi=True if providing a list of default values.")
        return multi

    @_log_call
    def build(self):
        full_options, default_value = get_options_and_default(options=self.options, multi=self.multi)
        return html.Div(
            [
                html.P(self.title) if self.title else None,
                dcc.Dropdown(
                    id=self.id,
                    options=full_options,
                    value=self.value if self.value is not None else default_value,
                    multi=self.multi,
                    persistence=True,
                    persistence_type="session",
                    className="selector_body_dropdown",
                ),
            ],
            className="selector_dropdown_container",
            id=f"{self.id}_outer",
        )



import pytest

from vizro import Vizro

# Allow our custom assert functions in tests_utils/asserts.py to do introspection nicely still.
# See https://pytest.org/en/7.4.x/how-to/assert.html#assertion-introspection-details
pytest.register_assert_rewrite("asserts")


@pytest.fixture(autouse=True)
def reset_managers():
    # this ensures that the managers are reset before and after each test
    # the reset BEFORE all tests is important because at pytest test collection, fixtures are evaluated and hence
    # the model_manager may be populated with models from other tests
    Vizro._reset()
    yield
    Vizro._reset()



from typing import List, NamedTuple, Optional, Tuple

import numpy as np
from numpy import ma

try:
    from pydantic.v1 import Field, PrivateAttr, ValidationError, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, ValidationError, validator

from vizro._constants import EMPTY_SPACE_CONST
from vizro.models import VizroBaseModel
from vizro.models._models_utils import get_unique_grid_component_ids

GAP_DEFAULT = "12px"
MIN_DEFAULT = "0px"


class ColRowGridLines(NamedTuple):
    """Stores CSS grid lines of grid area spanned by component."""

    col_start: int
    col_end: int
    row_start: int
    row_end: int


def _convert_to_combined_grid_coord(matrix: ma.MaskedArray) -> ColRowGridLines:
    """Converts matrix coordinates from user `grid` to one combined grid area spanned by component i.

    Required for validation of grid areas spanned by components.

    User-provided grid:  [[0, 1], [0, 2]]
    Matrix coordinates for component i=0: [(0, 0), (1, 0)]
    Grid coordinates for component i=0: ColRowGridLines(col_start=1, col_end=2, row_start=1, row_end=3)

    Args:
        matrix: Array that represents the user-provided grid with a mask on the relevant screen component i

    Returns:
        ColRowGridLines for combined area spanned by all placements of screen component i
    """
    matrix_coord = [(x, y) for x, row in enumerate(matrix) for y, value in enumerate(row) if ma.is_masked(value)]

    row_idx, col_idx = zip(*matrix_coord)
    return ColRowGridLines(
        col_start=min(col_idx) + 1, col_end=max(col_idx) + 2, row_start=min(row_idx) + 1, row_end=max(row_idx) + 2
    )


def _convert_to_single_grid_coord(matrix: ma.MaskedArray) -> List[ColRowGridLines]:
    """Converts matrix coordinates from user `grid` to list of grid areas spanned by each placement of component i.

    Required for validation of grid areas spanned by spaces, where the combined area does not need to be rectangular.

    User-provided grid:  [[0, 1], [0, 2]]
    Matrix coordinates for component i=0: [(0, 0), (1, 0)]
    Grid coordinates for component i=0: [ColRowGridLines(col_start=1, col_end=2, row_start=1, row_end=2),
                                         ColRowGridLines(col_start=1, col_end=2, row_start=2, row_end=3)]

    Args:
        matrix: Array that represents the user-provided grid with a mask on the relevant screen component i

    Returns:
        List of ColRowGridLines for each individual placement of screen component i
    """
    matrix_coord = [(x, y) for x, row in enumerate(matrix) for y, value in enumerate(row) if ma.is_masked(value)]

    return [
        ColRowGridLines(col_start=col_idx + 1, col_end=col_idx + 2, row_start=row_idx + 1, row_end=row_idx + 2)
        for row_idx, col_idx in matrix_coord
    ]


def _do_rectangles_overlap(r1: ColRowGridLines, r2: ColRowGridLines) -> bool:
    """Checks if rectangles r1 and r2 overlap in areas.

    1. Computes the min and max of r1 and r2 on both axes.
    2. Computes the boundaries of the intersection rectangle (x1=left, x2=right, y1=top, y2=bottom)
    3. Checks if the intersection is valid and has a positive non-zero area (x1 < x2 and y1 < y2)

    See: https://github.com/SFML/SFML/blob/12d81304e63e333174d943ba3ff572e38abd56e0/include/SFML/Graphics/Rect.inl#L109

    Args:
        r1: Tuple containing grid coordinates for screen component i
        r2: Tuple containing grid coordinates for screen component j

    Returns:
        Bool if rectangular grid area spanned by component i overlaps with the area of component j
    """
    x1 = max(min(r1.row_start, r1.row_end), min(r2.row_start, r2.row_end))
    y1 = max(min(r1.col_start, r1.col_end), min(r2.col_start, r2.col_end))
    x2 = min(max(r1.row_start, r1.row_end), max(r2.row_start, r2.row_end))
    y2 = min(max(r1.col_start, r1.col_end), max(r2.col_start, r2.col_end))
    return x1 < x2 and y1 < y2


def _validate_grid_areas(grid_areas: List[ColRowGridLines]) -> None:
    """Validates grid areas spanned by screen components in `Layout`."""
    for i, r1 in enumerate(grid_areas):
        for r2 in grid_areas[i + 1 :]:
            if _do_rectangles_overlap(r1, r2):
                raise ValueError("Grid areas must be rectangular and not overlap!")


def _get_grid_lines(grid: List[List[int]]) -> Tuple[List[ColRowGridLines], List[ColRowGridLines]]:
    """Gets list of ColRowGridLines for components and spaces on screen for validation and placement."""
    component_grid_lines = []
    unique_grid_idx = get_unique_grid_component_ids(grid)
    for component_idx in unique_grid_idx:
        matrix = ma.masked_equal(grid, component_idx)
        component_grid_lines.append(_convert_to_combined_grid_coord(matrix))

    matrix = ma.masked_equal(grid, EMPTY_SPACE_CONST)
    space_grid_lines = _convert_to_single_grid_coord(matrix=matrix)

    return component_grid_lines, space_grid_lines


class Layout(VizroBaseModel):
    """Grid specification to place chart/components on the [`Page`][vizro.models.Page].

    Args:
        grid (List[List[int]]): Grid specification to arrange components on screen.
        row_gap (str): Gap between rows in px. Defaults to `"12px"`.
        col_gap (str): Gap between columns in px. Defaults to `"12px"`.
        row_min_height (str): Minimum row height in px. Defaults to `"0px"`.
        col_min_width (str): Minimum column width in px. Defaults to `"0px"`.
    """

    grid: List[List[int]] = Field(..., description="Grid specification to arrange components on screen.")
    row_gap: str = Field(GAP_DEFAULT, description="Gap between rows in px. Defaults to 12px.", regex="[0-9]+px")
    col_gap: str = Field(GAP_DEFAULT, description="Gap between columns in px. Defaults to 12px.", regex="[0-9]+px")
    row_min_height: str = Field(MIN_DEFAULT, description="Minimum row height in px. Defaults to 0px.", regex="[0-9]+px")
    col_min_width: str = Field(
        MIN_DEFAULT, description="Minimum column width in px. Defaults to 0px.", regex="[0-9]+px"
    )
    _component_grid_lines: Optional[List[ColRowGridLines]] = PrivateAttr()

    @validator("grid")
    def validate_grid(cls, grid):
        if len({len(row) for row in grid}) > 1:
            raise ValueError("All rows must be of same length.")

        # Validate grid type and values
        unique_grid_idx = get_unique_grid_component_ids(grid)
        if 0 not in unique_grid_idx or not np.array_equal(unique_grid_idx, np.arange((unique_grid_idx.max() + 1))):
            raise ValueError("Grid must contain consecutive integers starting from 0.")

        # Validates grid areas spanned by components and spaces
        component_grid_lines, space_grid_lines = _get_grid_lines(grid)
        _validate_grid_areas(component_grid_lines + space_grid_lines)
        return grid

    def __init__(self, **data):
        super().__init__(**data)
        self._component_grid_lines = _get_grid_lines(self.grid)[0]

    @property
    def component_grid_lines(self):
        return self._component_grid_lines


if __name__ == "__main__":
    print(repr(Layout(grid=[[0, 1], [0, 2]])))  # noqa: T201

    try:
        Layout(grid=[[0, 1], [1, 0]])
    except ValidationError as e:
        print(e)  # noqa: T201



"""The action loop creates all the required action callbacks and its components."""


from dash import html

from vizro.actions._action_loop._action_loop_utils import _get_actions_on_registered_pages
from vizro.actions._action_loop._build_action_loop_callbacks import _build_action_loop_callbacks
from vizro.actions._action_loop._get_action_loop_components import _get_action_loop_components


class ActionLoop:
    @classmethod
    def _create_app_callbacks(cls) -> html.Div:
        """Builds callbacks for the action loop and for each Action in the Dashboard and returns their components.

        Returns:
            List of required components for the action loop and for each `Action` in the `Dashboard`.
        """
        return html.Div([cls._build_action_loop(), cls._build_actions_models()], id="app_components_div", hidden=True)

    @staticmethod
    def _build_action_loop():
        """Builds callbacks for the action loop and returns required components for the action loop mechanism to work.

        Returns:
            List of required components for the action loop e.g. List[dcc.Store, html.Div].
        """
        _build_action_loop_callbacks()
        return _get_action_loop_components()

    @staticmethod
    def _build_actions_models():
        """Builds a callback for each `Action` model and returns required components for these callbacks.

        Returns:
            List of required components for each `Action` in the `Dashboard` e.g. List[dcc.Download]
        """
        actions = _get_actions_on_registered_pages()
        return html.Div([action.build() for action in actions], id="app_action_models_components_div", hidden=True)



import logging
from typing import List, Literal

from dash import ctx, dcc
from dash.exceptions import MissingCallbackContextException
from plotly import graph_objects as go

try:
    from pydantic.v1 import Field, PrivateAttr, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, validator

import vizro.plotly.express as px
from vizro import _themes as themes
from vizro.managers import data_manager
from vizro.models import Action, VizroBaseModel
from vizro.models._action._actions_chain import _action_validator_factory
from vizro.models._components._components_utils import _process_callable_data_frame
from vizro.models._models_utils import _log_call
from vizro.models.types import CapturedCallable

logger = logging.getLogger(__name__)


class Graph(VizroBaseModel):
    """Wrapper for `dcc.Graph` to visualize charts in dashboard.

    Args:
        type (Literal["graph"]): Defaults to `"graph"`.
        figure (CapturedCallable): See [`CapturedCallable`][vizro.models.types.CapturedCallable].
        actions (List[Action]): See [`Action`][vizro.models.Action]. Defaults to `[]`.
    """

    type: Literal["graph"] = "graph"
    figure: CapturedCallable = Field(..., import_path=px)
    actions: List[Action] = []

    # Component properties for actions and interactions
    _output_property: str = PrivateAttr("figure")

    # Re-used validators
    _set_actions = _action_validator_factory("clickData")
    _validate_callable = validator("figure", allow_reuse=True)(_process_callable_data_frame)

    # Convenience wrapper/syntactic sugar.
    def __call__(self, **kwargs):
        kwargs.setdefault("data_frame", data_manager._get_component_data(str(self.id)))
        fig = self.figure(**kwargs)

        # Remove top margin if title is provided
        if fig.layout.title.text is None:
            fig.update_layout(margin_t=24)

        # Possibly we should enforce that __call__ can only be used within the context of a callback, but it's easy
        # to just swallow up the error here as it doesn't cause any problems.
        try:
            # At the moment theme_selector is always present so this if statement is redundant, but possibly in
            # future we'll have callbacks that do Graph.__call__() without theme_selector set.
            if "theme_selector" in ctx.args_grouping.get("external", {}):
                fig = self._update_theme(fig, ctx.args_grouping["external"]["theme_selector"]["value"])
        except MissingCallbackContextException:
            logger.info("fig.update_layout called outside of callback context.")
        return fig

    # Convenience wrapper/syntactic sugar.
    def __getitem__(self, arg_name: str):
        # pydantic discriminated union validation seems to try Graph["type"], which throws an error unless we
        # explicitly redirect it to the correct attribute.
        if arg_name == "type":
            return self.type
        return self.figure[arg_name]

    @_log_call
    def build(self):
        # The empty figure here is just a placeholder designed to be replaced by the actual figure when the filters
        # etc. are applied. It only appears on the screen for a brief instant, but we need to make sure it's
        # transparent and has no axes so it doesn't draw anything on the screen which would flicker away when the
        # graph callback is executed to make the dcc.Loading icon appear.
        return dcc.Loading(
            dcc.Graph(
                id=self.id,
                figure=go.Figure(
                    layout={
                        "paper_bgcolor": "rgba(0,0,0,0)",
                        "plot_bgcolor": "rgba(0,0,0,0)",
                        "xaxis": {"visible": False},
                        "yaxis": {"visible": False},
                    }
                ),
                config={
                    "autosizable": True,
                    "frameMargins": 0,
                    "responsive": True,
                },
                className="chart_container",
            ),
            color="grey",
            parent_className="loading-container",
        )

    @staticmethod
    def _update_theme(fig: go.Figure, theme_selector: bool):
        # Basically the same as doing fig.update_layout(template="vizro_light/dark") but works for both the call in
        # self.__call__ and in the update_graph_theme callback.
        fig["layout"]["template"] = themes.dark if theme_selector else themes.light
        return fig



"""Unit tests for vizro.actions._callback_mapping._get_action_callback_mapping file."""

import json

import dash
import plotly
import pytest

import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro
from vizro.actions import export_data, filter_interaction
from vizro.actions._callback_mapping._get_action_callback_mapping import _get_action_callback_mapping
from vizro.models.types import capture


@capture("action")
def custom_action_example():
    pass


# custom action with same name as some predefined action
def get_custom_action_with_known_name():
    @capture("action")
    def export_data():
        pass

    return export_data()


@pytest.fixture
def managers_one_page_four_controls_three_figures_filter_interaction(request, dash_data_table_with_id):
    """Instantiates managers with one page that contains four controls, two graphs and filter interaction."""
    # If the fixture is parametrised set the targets. Otherwise, set export_data without targets.
    export_data_action_function = export_data(targets=request.param) if hasattr(request, "param") else export_data()

    vm.Page(
        id="test_page",
        title="My first dashboard",
        components=[
            vm.Graph(
                id="scatter_chart",
                figure=px.scatter(px.data.gapminder(), x="lifeExp", y="gdpPercap", custom_data=["continent"]),
                actions=[
                    vm.Action(id="filter_interaction_action", function=filter_interaction(targets=["scatter_chart_2"]))
                ],
            ),
            vm.Graph(
                id="scatter_chart_2",
                figure=px.scatter(px.data.gapminder(), x="lifeExp", y="gdpPercap", custom_data=["continent"]),
                actions=[vm.Action(id="custom_action", function=custom_action_example())],
            ),
            vm.Table(
                id="vizro_table",
                figure=dash_data_table_with_id,
                actions=[
                    vm.Action(
                        id="table_filter_interaction_action",
                        function=filter_interaction(targets=["scatter_chart", "scatter_chart_2"]),
                    )
                ],
            ),
            vm.Button(
                id="export_data_button",
                actions=[
                    vm.Action(id="export_data_action", function=export_data_action_function),
                    vm.Action(id="export_data_custom_action", function=get_custom_action_with_known_name()),
                ],
            ),
        ],
        controls=[
            vm.Filter(id="filter_continent", column="continent", selector=vm.Dropdown(id="filter_continent_selector")),
            vm.Filter(id="filter_country", column="country", selector=vm.Dropdown(id="filter_country_selector")),
            vm.Parameter(
                id="parameter_x",
                targets=["scatter_chart.x", "scatter_chart_2.x"],
                selector=vm.Dropdown(
                    id="parameter_x_selector",
                    options=["lifeExp", "gdpPercap", "pop"],
                    multi=False,
                    value="gdpPercap",
                ),
            ),
            vm.Parameter(
                id="parameter_y",
                targets=["scatter_chart.y", "scatter_chart_2.y"],
                selector=vm.Dropdown(
                    id="parameter_y_selector",
                    options=["lifeExp", "gdpPercap", "pop"],
                    multi=False,
                    value="lifeExp",
                ),
            ),
            vm.Parameter(
                id="vizro_table_row_selectable",
                targets=["vizro_table.row_selectable"],
                selector=vm.Dropdown(
                    id="parameter_table_row_selectable",
                    options=["multi", "single"],
                    multi=False,
                    value="single",
                ),
            ),
        ],
    )
    Vizro._pre_build()


@pytest.fixture
def action_callback_inputs_expected():
    return {
        "filters": [
            dash.State("filter_continent_selector", "value"),
            dash.State("filter_country_selector", "value"),
        ],
        "parameters": [
            dash.State("parameter_x_selector", "value"),
            dash.State("parameter_y_selector", "value"),
            dash.State("parameter_table_row_selectable", "value"),
        ],
        "filter_interaction": [
            {
                "clickData": dash.State("scatter_chart", "clickData"),
            },
            {
                "active_cell": dash.State("underlying_table_id", "active_cell"),
                "derived_viewport_data": dash.State("underlying_table_id", "derived_viewport_data"),
            },
        ],
        "theme_selector": dash.State("theme_selector", "on"),
    }


@pytest.fixture
def action_callback_outputs_expected(request):
    targets = request.param
    return {
        target["component_id"]: dash.Output(target["component_id"], target["component_property"]) for target in targets
    }


@pytest.fixture
def export_data_inputs_expected():
    return {
        "filters": [
            dash.State("filter_continent_selector", "value"),
            dash.State("filter_country_selector", "value"),
        ],
        "parameters": [],
        "filter_interaction": [
            {
                "clickData": dash.State("scatter_chart", "clickData"),
            },
            {
                "active_cell": dash.State("underlying_table_id", "active_cell"),
                "derived_viewport_data": dash.State("underlying_table_id", "derived_viewport_data"),
            },
        ],
        "theme_selector": [],
    }


@pytest.fixture
def export_data_outputs_expected(request):
    return {
        f"download_dataframe_{target}": dash.Output(
            {"action_id": "export_data_action", "target_id": target, "type": "download_dataframe"}, "data"
        )
        for target in request.param
    }


@pytest.fixture
def export_data_components_expected(request):
    return [
        dash.dcc.Download(id={"type": "download_dataframe", "action_id": "export_data_action", "target_id": target})
        for target in request.param
    ]


@pytest.mark.usefixtures("managers_one_page_four_controls_three_figures_filter_interaction")
class TestCallbackMapping:
    """Tests action callback mapping for predefined and custom actions."""

    @pytest.mark.parametrize(
        "action_id, callback_mapping_inputs_expected",
        [
            ("filter_action_filter_continent", "action_callback_inputs_expected"),
            ("filter_interaction_action", "action_callback_inputs_expected"),
            ("parameter_action_parameter_x", "action_callback_inputs_expected"),
            ("on_page_load_action_action_test_page", "action_callback_inputs_expected"),
            ("export_data_action", "export_data_inputs_expected"),
        ],
    )
    def test_action_callback_mapping_inputs(self, action_id, callback_mapping_inputs_expected, request):
        result = _get_action_callback_mapping(
            action_id=action_id,
            argument="inputs",
        )

        callback_mapping_inputs_expected = request.getfixturevalue(callback_mapping_inputs_expected)
        assert result == callback_mapping_inputs_expected

    @pytest.mark.parametrize(
        "action_id, action_callback_outputs_expected",
        [
            (
                "filter_action_filter_continent",
                [
                    {"component_id": "scatter_chart", "component_property": "figure"},
                    {"component_id": "scatter_chart_2", "component_property": "figure"},
                    {"component_id": "vizro_table", "component_property": "children"},
                ],
            ),
            ("filter_interaction_action", [{"component_id": "scatter_chart_2", "component_property": "figure"}]),
            (
                "table_filter_interaction_action",
                [
                    {"component_id": "scatter_chart", "component_property": "figure"},
                    {"component_id": "scatter_chart_2", "component_property": "figure"},
                ],
            ),
            (
                "parameter_action_parameter_x",
                [
                    {"component_id": "scatter_chart", "component_property": "figure"},
                    {"component_id": "scatter_chart_2", "component_property": "figure"},
                ],
            ),
            (
                "parameter_action_parameter_y",
                [
                    {"component_id": "scatter_chart", "component_property": "figure"},
                    {"component_id": "scatter_chart_2", "component_property": "figure"},
                ],
            ),
            (
                "on_page_load_action_action_test_page",
                [
                    {"component_id": "scatter_chart", "component_property": "figure"},
                    {"component_id": "scatter_chart_2", "component_property": "figure"},
                    {"component_id": "vizro_table", "component_property": "children"},
                ],
            ),
            (
                "parameter_action_vizro_table_row_selectable",
                [
                    {"component_id": "vizro_table", "component_property": "children"},
                ],
            ),
        ],
        indirect=["action_callback_outputs_expected"],
    )
    def test_action_callback_mapping_outputs(self, action_id, action_callback_outputs_expected):
        result = _get_action_callback_mapping(
            action_id=action_id,
            argument="outputs",
        )
        assert result == action_callback_outputs_expected

    @pytest.mark.parametrize(
        "export_data_outputs_expected",
        [("scatter_chart", "scatter_chart_2", "vizro_table")],
        indirect=True,
    )
    def test_export_data_no_targets_set_mapping_outputs(self, export_data_outputs_expected):
        result = _get_action_callback_mapping(
            action_id="export_data_action",
            argument="outputs",
        )

        assert result == export_data_outputs_expected

    @pytest.mark.parametrize(
        "managers_one_page_four_controls_three_figures_filter_interaction, export_data_outputs_expected",
        [
            (None, ["scatter_chart", "scatter_chart_2", "vizro_table"]),
            ([], ["scatter_chart", "scatter_chart_2", "vizro_table"]),
            (["scatter_chart"], ["scatter_chart"]),
            (["scatter_chart", "scatter_chart_2"], ["scatter_chart", "scatter_chart_2"]),
        ],
        indirect=True,
    )
    def test_export_data_targets_set_mapping_outputs(
        self, managers_one_page_four_controls_three_figures_filter_interaction, export_data_outputs_expected
    ):
        result = _get_action_callback_mapping(
            action_id="export_data_action",
            argument="outputs",
        )

        assert result == export_data_outputs_expected

    @pytest.mark.parametrize(
        "export_data_components_expected",
        [("scatter_chart", "scatter_chart_2", "vizro_table")],
        indirect=True,
    )
    def test_export_data_no_targets_set_mapping_components(self, export_data_components_expected):
        result_components = _get_action_callback_mapping(
            action_id="export_data_action",
            argument="components",
        )

        result = json.dumps(result_components, cls=plotly.utils.PlotlyJSONEncoder)
        expected = json.dumps(export_data_components_expected, cls=plotly.utils.PlotlyJSONEncoder)
        assert result == expected

    @pytest.mark.parametrize(
        "managers_one_page_four_controls_three_figures_filter_interaction, export_data_components_expected",
        [
            (None, ["scatter_chart", "scatter_chart_2", "vizro_table"]),
            ([], ["scatter_chart", "scatter_chart_2", "vizro_table"]),
            (["scatter_chart"], ["scatter_chart"]),
            (["scatter_chart", "scatter_chart_2"], ["scatter_chart", "scatter_chart_2"]),
        ],
        indirect=True,
    )
    def test_export_data_targets_set_mapping_components(
        self, managers_one_page_four_controls_three_figures_filter_interaction, export_data_components_expected
    ):
        result_components = _get_action_callback_mapping(
            action_id="export_data_action",
            argument="components",
        )
        result = json.dumps(result_components, cls=plotly.utils.PlotlyJSONEncoder)
        expected = json.dumps(export_data_components_expected, cls=plotly.utils.PlotlyJSONEncoder)
        assert result == expected

    def test_known_action_unknown_argument(self):
        result = _get_action_callback_mapping(
            action_id="export_data_action",
            argument="unknown-argument",
        )
        assert result == {}

    # "export_data_custom_action" represents a unique scenario within custom actions, where the function's name
    # coincides with an already predefined action function (in this instance, "export_data").
    # It requires handling them as conventional custom actions.
    @pytest.mark.parametrize("action_id", ["custom_action", "export_data_custom_action"])
    @pytest.mark.parametrize(
        "argument, expected",
        [
            ("inputs", {}),
            ("outputs", {}),
            ("components", []),
            ("unknown-argument", {}),
        ],
    )
    def test_custom_action_mapping(self, action_id, argument, expected):
        result = _get_action_callback_mapping(
            action_id=action_id,
            argument=argument,
        )
        assert result == expected



import plotly.express as px

import vizro.plotly.express as hpx


def test_non_chart_unchanged():
    assert hpx.data is px.data


def test_chart_wrapped():
    graph = hpx.scatter(px.data.iris(), x="petal_width", y="petal_length")
    assert graph._captured_callable._function is px.scatter
    assert hpx.scatter is not px.scatter



import pytest
from asserts import assert_component_equal

import vizro.models as vm
from vizro import Vizro
from vizro.managers import model_manager


@pytest.fixture
def dashboard_result(request):
    # Inject the navigation into the dashboard. Note we need to call request.param since they are all lambda functions.
    dashboard = vm.Dashboard(
        pages=[
            vm.Page(title="Page 1", components=[vm.Button()]),
            vm.Page(title="Page 2", components=[vm.Button()]),
        ],
        navigation=request.param(),
    )
    Vizro()._pre_build()
    # Ideally we would do Vizro._reset after yield dashboard, but that won't work since the fixture is used
    # indirectly and needs to be reset immediately after it's used rather than after running the test.
    # We can't do Vizro._reset since current implementation of navigation uses dash.page_registry to look up a page,
    # so that must remain populated.
    model_manager._clear()
    return dashboard


# Looks hacky but this is the easiest way to use the fixture twice in the same test.
dashboard_expected = dashboard_result

# All the cases need to built lazily - instantiating them directly would not raise validation errors that the
# specified page cannot be found in the page registry.
# In all test cases, the first lambda should be thought of as the "input" configuration, and the second is the
# configuration that results after we've filled in default values etc.
# fmt: off
accordion_cases = [
    (
        lambda: None,
        lambda: vm.Navigation(nav_selector=vm.Accordion(pages=["Page 1", "Page 2"])),
    ),
    (
        lambda: vm.Navigation(),
        lambda: vm.Navigation(nav_selector=vm.Accordion(pages=["Page 1", "Page 2"])),
    ),
    (
        lambda: vm.Navigation(nav_selector=vm.Accordion()),
        lambda: vm.Navigation(nav_selector=vm.Accordion(pages=["Page 1", "Page 2"])),
    ),
    (
        lambda: vm.Navigation(pages=["Page 1"]),
        lambda: vm.Navigation(nav_selector=vm.Accordion(pages=["Page 1"]))),
    (
        lambda: vm.Navigation(pages=["Page 1"], nav_selector=vm.Accordion(pages=["Page 2"])),
        lambda: vm.Navigation(nav_selector=vm.Accordion(pages=["Page 2"])),
    ),
    (
        lambda: vm.Navigation(pages={"Group 1": ["Page 1"], "Group 2": ["Page 2"]}),
        lambda: vm.Navigation(nav_selector=vm.Accordion(pages={"Group 1": ["Page 1"], "Group 2": ["Page 2"]})),
    ),
]
# fmt: on

navbar_flat_cases = [
    (
        lambda: vm.Navigation(nav_selector=vm.NavBar()),
        lambda: vm.Navigation(nav_selector=vm.NavBar(pages=["Page 1", "Page 2"])),
    ),
    (
        lambda: vm.Navigation(nav_selector=vm.NavBar()),
        lambda: vm.Navigation(
            nav_selector=vm.NavBar(
                items=[
                    vm.NavLink(label="Page 1", pages=["Page 1"], icon="filter_1"),
                    vm.NavLink(label="Page 2", pages=["Page 2"], icon="filter_2"),
                ]
            )
        ),
    ),
    (
        lambda: vm.Navigation(pages=["Page 1"], nav_selector=vm.NavBar()),
        lambda: vm.Navigation(nav_selector=vm.NavBar(pages=["Page 1"])),
    ),
    (
        lambda: vm.Navigation(pages=["Page 1"], nav_selector=vm.NavBar(pages=["Page 2"])),
        lambda: vm.Navigation(nav_selector=vm.NavBar(pages=["Page 2"])),
    ),
    (
        lambda: vm.Navigation(pages=["Page 1"], nav_selector=vm.NavBar()),
        lambda: vm.Navigation(
            nav_selector=vm.NavBar(
                items=[
                    vm.NavLink(label="Page 1", pages=["Page 1"], icon="filter_1"),
                ]
            )
        ),
    ),
]

navbar_grouped_cases = [
    (
        lambda: vm.Navigation(pages={"Group 1": ["Page 1"], "Group 2": ["Page 2"]}, nav_selector=vm.NavBar()),
        lambda: vm.Navigation(nav_selector=vm.NavBar(pages={"Group 1": ["Page 1"], "Group 2": ["Page 2"]})),
    ),
    (
        lambda: vm.Navigation(pages={"Group 1": ["Page 1"], "Group 2": ["Page 2"]}, nav_selector=vm.NavBar()),
        lambda: vm.Navigation(
            nav_selector=vm.NavBar(
                items=[
                    vm.NavLink(label="Group 1", pages=["Page 1"], icon="filter_1"),
                    vm.NavLink(label="Group 2", pages=["Page 2"], icon="filter_2"),
                ]
            )
        ),
    ),
    (
        lambda: vm.Navigation(
            nav_selector=vm.NavBar(
                items=[
                    vm.NavLink(label="Group 1", pages={"Subgroup 1": ["Page 1"]}, icon="dashboard"),
                    vm.NavLink(label="Group 2", pages={"Subgroup 2": ["Page 2"]}),
                ]
            )
        ),
        lambda: vm.Navigation(
            nav_selector=vm.NavBar(
                items=[
                    vm.NavLink(label="Group 1", pages={"Subgroup 1": ["Page 1"]}, icon="dashboard"),
                    vm.NavLink(label="Group 2", pages={"Subgroup 2": ["Page 2"]}, icon="filter_2"),
                ]
            )
        ),
    ),
]


# Label the test cases to make it easier to see which one failed.
def label_cases(cases, label):
    return [pytest.param(*case, id=f"{label}-{case_id}") for case_id, case in enumerate(cases)]


accordion_cases = label_cases(accordion_cases, "accordion")
navbar_flat_cases = label_cases(navbar_flat_cases, "navbar_flat")
navbar_grouped_cases = label_cases(navbar_grouped_cases, "navbar_grouped")


@pytest.mark.parametrize(
    "dashboard_result, dashboard_expected",
    accordion_cases + navbar_flat_cases + navbar_grouped_cases,
    indirect=True,
)
def test_navigation_build(dashboard_result, dashboard_expected):
    result = dashboard_result.navigation.build()
    expected = dashboard_expected.navigation.build()
    assert_component_equal(result, expected, keys_to_strip={"id"})



from typing import Literal

import dash_bootstrap_components as dbc
from dash import dcc, get_relative_path, html

try:
    from pydantic.v1 import Field
except ImportError:  # pragma: no cov
    from pydantic import Field

from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call


class Card(VizroBaseModel):
    """Creates a card utilizing `dcc.Markdown` as title and text component.

    Args:
        type (Literal["card"]): Defaults to `"card"`.
        text (str): Markdown string to create card title/text that should adhere to the CommonMark Spec.
        href (str): URL (relative or absolute) to navigate to. If not provided the Card serves as a text card
            only. Defaults to `""`.
    """

    type: Literal["card"] = "card"
    text: str = Field(
        ..., description="Markdown string to create card title/text that should adhere to the CommonMark Spec."
    )
    href: str = Field(
        "",
        description="URL (relative or absolute) to navigate to. If not provided the Card serves as a text card only.",
    )

    @_log_call
    def build(self):
        text = dcc.Markdown(self.text, className="card_text", dangerously_allow_html=False, id=self.id)
        button = html.Div(
            dbc.Button(
                href=get_relative_path(self.href) if self.href.startswith("/") else self.href,
                className="card_button",
            ),
            className="button_container",
        )

        card_container = "nav_card_container" if self.href else "card_container"

        return html.Div([text, button if self.href else None], className=card_container, id=f"{self.id}_outer")



"""Fixtures to be shared across several tests."""

import plotly.graph_objects as go
import pytest

import vizro.models as vm
import vizro.plotly.express as px
from vizro import Vizro
from vizro.tables import dash_data_table


@pytest.fixture
def gapminder():
    return px.data.gapminder()


@pytest.fixture
def standard_px_chart(gapminder):
    return px.scatter(
        data_frame=gapminder,
        x="gdpPercap",
        y="lifeExp",
        size="pop",
        color="continent",
        hover_name="country",
        size_max=60,
    )


@pytest.fixture
def standard_dash_table(gapminder):
    return dash_data_table(data_frame=gapminder)


@pytest.fixture
def dash_data_table_with_id(gapminder):
    return dash_data_table(id="underlying_table_id", data_frame=gapminder)


@pytest.fixture
def standard_go_chart(gapminder):
    return go.Figure(data=go.Scatter(x=gapminder["gdpPercap"], y=gapminder["lifeExp"], mode="markers"))


@pytest.fixture()
def page_1():
    return vm.Page(title="Page 1", components=[vm.Button()])


@pytest.fixture()
def page_2():
    return vm.Page(title="Page 2", components=[vm.Button()])


@pytest.fixture()
def vizro_app():
    """Fixture to instantiate Vizro/Dash app.

    Required if Vizro._pre_build or dashboard.pre_build is called since dash.register_page can only be called after
    app instantiation.pages.
    """
    return Vizro()



from __future__ import annotations

import itertools

import dash
import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
from dash import html

try:
    from pydantic.v1 import Field, PrivateAttr, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, PrivateAttr, validator

from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models._navigation._navigation_utils import _validate_pages
from vizro.models._navigation.accordion import Accordion
from vizro.models.types import NavPagesType


class NavLink(VizroBaseModel):
    """Icon to be used in Navigation Bar of Dashboard.

    Args:
        pages (NavPagesType): See [`NavPagesType`][vizro.models.types.NavPagesType]. Defaults to `[]`.
        label (str): Text description of the icon for use in tooltip.
        icon (str): Icon name from [Google Material icons library](https://fonts.google.com/icons). Defaults to `""`.

    """

    pages: NavPagesType = []
    label: str = Field(..., description="Text description of the icon for use in tooltip.")
    icon: str = Field("", description="Icon name from Google Material icons library.")

    _nav_selector: Accordion = PrivateAttr()

    # Re-used validators
    _validate_pages = validator("pages", allow_reuse=True)(_validate_pages)

    @validator("icon")
    def validate_icon(cls, icon) -> str:
        return icon.strip().lower().replace(" ", "_")

    @_log_call
    def pre_build(self):
        from vizro.models._navigation.accordion import Accordion

        self._nav_selector = Accordion(pages=self.pages)

    @_log_call
    def build(self, *, active_page_id=None):
        # _nav_selector is an Accordion, so _nav_selector._pages is guaranteed to be Dict[str, List[str]].
        all_page_ids = list(itertools.chain(*self._nav_selector.pages.values()))
        first_page_id = all_page_ids[0]
        item_active = active_page_id in all_page_ids

        try:
            first_page = dash.page_registry[first_page_id]
        except KeyError as exc:
            raise KeyError(
                f"Page with ID {first_page_id} cannot be found. Please add the page to `Dashboard.pages`"
            ) from exc

        button = dbc.Button(
            [
                dmc.Tooltip(
                    label=self.label,
                    offset=4,
                    withArrow=True,
                    children=[html.Span(self.icon, className="material-symbols-outlined")],
                    position="bottom-start",
                )
            ],
            id=self.id,
            className="icon-button",
            href=first_page["relative_path"],
            active=item_active,
        )

        # Only build the nav_selector (id="nav-panel") if the item is active.
        if item_active:
            return html.Div([button, self._nav_selector.build(active_page_id=active_page_id)])

        return html.Div(button)



from functools import partial
from typing import Any, Dict, List, NamedTuple

try:
    from pydantic.v1 import validator
except ImportError:  # pragma: no cov
    from pydantic import validator

from vizro.models import Action, VizroBaseModel


class Trigger(NamedTuple):
    component_id: str
    component_property: str


class ActionsChain(VizroBaseModel):
    trigger: Trigger
    actions: List[Action] = []


# Validators for reuse in other models to convert to ActionsChain
def _set_actions(actions: List[Action], values: Dict[str, Any], trigger_property: str) -> List[ActionsChain]:
    return [
        ActionsChain(
            trigger=Trigger(
                component_id=values["id"],
                component_property=trigger_property,
            ),
            actions=actions,
        )
    ]


def _action_validator_factory(trigger_property: str):
    set_actions = partial(_set_actions, trigger_property=trigger_property)
    return validator("actions", allow_reuse=True)(set_actions)



"""Contains utilities to extract the Action and ActionsChain models from registered pages only."""

from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, List

from dash import page_registry

from vizro.managers import model_manager
from vizro.models import VizroBaseModel

if TYPE_CHECKING:
    from vizro.models import Action, Page
    from vizro.models._action._actions_chain import ActionsChain


def _get_actions(model: VizroBaseModel) -> List[ActionsChain]:
    """Gets the list of the ActionsChain models for any VizroBaseModel model."""
    if hasattr(model, "selector"):
        return model.selector.actions
    elif hasattr(model, "actions"):
        return model.actions
    return []


def _get_all_actions_chains_on_page(page: Page) -> List[ActionsChain]:
    """Gets the list of the ActionsChain models for the Page model."""
    return [
        actions_chain
        for page_item in chain([page], page.components, page.controls)
        for actions_chain in _get_actions(model=page_item)
    ]


def _get_actions_chains_on_registered_pages() -> List[ActionsChain]:
    """Gets list of ActionsChain models for registered pages."""
    actions_chains: List[ActionsChain] = []
    for registered_page in page_registry.values():
        try:
            page: Page = model_manager[registered_page["module"]]
        except KeyError:
            continue
        actions_chains.extend(_get_all_actions_chains_on_page(page=page))
    return actions_chains


def _get_actions_on_registered_pages() -> List[Action]:
    """Gets list of Action models for registered pages."""
    return [action for action_chain in _get_actions_chains_on_registered_pages() for action in action_chain.actions]



"""Unit tests for vizro.models.NavBar."""
import re

import dash_bootstrap_components as dbc
import dash_mantine_components as dmc
import pytest
from asserts import STRIP_ALL, assert_component_equal
from dash import html

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm

pytestmark = pytest.mark.usefixtures("prebuilt_two_page_dashboard")


class TestNavBarInstantiation:
    """Tests NavBar model instantiation."""

    def test_nav_bar_mandatory_only(self):
        nav_bar = vm.NavBar()

        assert hasattr(nav_bar, "id")
        assert nav_bar.pages == {}
        assert nav_bar.items == []

    def test_nav_bar_mandatory_and_optional(self, pages_as_dict):
        nav_link = vm.NavLink(label="Label")
        nav_bar = vm.NavBar(id="nav-bar", pages=pages_as_dict, items=[nav_link])

        assert nav_bar.id == "nav-bar"
        assert nav_bar.pages == pages_as_dict
        assert nav_bar.items == [nav_link]

    def test_valid_pages_as_list(self, pages_as_list):
        nav_bar = vm.NavBar(pages=pages_as_list)
        assert nav_bar.pages == {"Page 1": ["Page 1"], "Page 2": ["Page 2"]}

    @pytest.mark.parametrize("pages", [{"Group": []}, []])
    def test_invalid_field_pages_no_ids_provided(self, pages):
        with pytest.raises(ValidationError, match="Ensure this value has at least 1 item."):
            vm.NavBar(pages=pages)

    def test_invalid_field_pages_wrong_input_type(self):
        with pytest.raises(ValidationError, match="unhashable type: 'Page'"):
            vm.NavBar(pages=[vm.Page(title="Page 3", components=[vm.Button()])])

    @pytest.mark.parametrize("pages", [["non existent page"], {"Group": ["non existent page"]}])
    def test_invalid_page(self, pages):
        with pytest.raises(
            ValidationError, match=re.escape("Unknown page ID ['non existent page'] provided to argument 'pages'.")
        ):
            vm.NavBar(pages=pages)


class TestNavBarPreBuildMethod:
    def test_default_items(self, pages_as_dict):
        nav_bar = vm.NavBar(pages=pages_as_dict)
        nav_bar.pre_build()
        assert all(isinstance(item, vm.NavLink) for item in nav_bar.items)
        assert all(item.icon == f"filter_{position}" for position, item in enumerate(nav_bar.items, 1))

    def test_items_with_with_pages_icons(self, pages_as_dict):
        nav_links = [
            vm.NavLink(label="Label", pages={"Group 1": ["Page 1"]}, icon="Home"),
            vm.NavLink(label="Label", pages={"Group 2": ["Page 2"]}),
        ]
        nav_bar = vm.NavBar(pages=pages_as_dict, items=nav_links)
        nav_bar.pre_build()
        assert nav_bar.items == nav_links
        assert nav_bar.items[0].icon == "home"
        assert nav_bar.items[1].icon == "filter_2"


class TestNavBarBuildMethod:
    """Tests NavBar model build method."""

    common_args = {"offset": 4, "withArrow": True, "position": "bottom-start"}

    def test_nav_bar_active_pages_as_dict(self, pages_as_dict):
        nav_bar = vm.NavBar(pages=pages_as_dict)
        nav_bar.pre_build()
        built_nav_bar = nav_bar.build(active_page_id="Page 1")
        expected_button = html.Div(
            [
                dbc.Button(
                    children=[dmc.Tooltip(label="Group", children=[html.Span("filter_1")], **self.common_args)],
                    active=True,
                    href="/",
                )
            ]
        )
        assert_component_equal(built_nav_bar["nav-bar"], expected_button, keys_to_strip={"id", "className"})
        assert_component_equal(built_nav_bar["nav-panel"], html.Div(id="nav-panel"), keys_to_strip={"children"})
        assert_component_equal(built_nav_bar["nav-panel"].children, [dbc.Accordion()], keys_to_strip=STRIP_ALL)

    def test_nav_bar_active_pages_as_list(self, pages_as_list):
        nav_bar = vm.NavBar(pages=pages_as_list)
        nav_bar.pre_build()
        built_nav_bar = nav_bar.build(active_page_id="Page 1")
        expected_buttons = html.Div(
            [
                dbc.Button(
                    children=[dmc.Tooltip(label="Page 1", children=[html.Span("filter_1")], **self.common_args)],
                    active=True,
                    href="/",
                ),
                dbc.Button(
                    children=[dmc.Tooltip(label="Page 2", children=[html.Span("filter_2")], **self.common_args)],
                    active=False,
                    href="/page-2",
                ),
            ]
        )
        assert_component_equal(built_nav_bar["nav-bar"], expected_buttons, keys_to_strip={"id", "className"})
        assert_component_equal(built_nav_bar["nav-panel"], html.Div(id="nav-panel", hidden=True))

    def test_nav_bar_not_active_pages_as_dict(self, pages_as_dict):
        nav_bar = vm.NavBar(pages=pages_as_dict)
        nav_bar.pre_build()
        built_nav_bar = nav_bar.build(active_page_id="Page 3")
        expected_button = html.Div(
            [
                dbc.Button(
                    children=[dmc.Tooltip(label="Group", children=[html.Span("filter_1")], **self.common_args)],
                    active=False,
                    href="/",
                )
            ]
        )
        assert_component_equal(built_nav_bar["nav-bar"], expected_button, keys_to_strip={"id", "className"})
        assert_component_equal(built_nav_bar["nav-panel"], html.Div(hidden=True, id="nav-panel"))

    def test_nav_bar_not_active_pages_as_list(self, pages_as_list):
        nav_bar = vm.NavBar(pages=pages_as_list)
        nav_bar.pre_build()
        built_nav_bar = nav_bar.build(active_page_id="Page 3")
        expected_buttons = html.Div(
            [
                dbc.Button(
                    children=[dmc.Tooltip(label="Page 1", children=[html.Span("filter_1")], **self.common_args)],
                    active=False,
                    href="/",
                ),
                dbc.Button(
                    children=[dmc.Tooltip(label="Page 2", children=[html.Span("filter_2")], **self.common_args)],
                    active=False,
                    href="/page-2",
                ),
            ]
        )
        assert_component_equal(built_nav_bar["nav-bar"], expected_buttons, keys_to_strip={"id", "className"})
        assert_component_equal(built_nav_bar["nav-panel"], html.Div(id="nav-panel", hidden=True))



import itertools
from collections.abc import Mapping
from typing import Dict, List, Literal

import dash
import dash_bootstrap_components as dbc
from dash import html

try:
    from pydantic.v1 import Field, validator
except ImportError:  # pragma: no cov
    from pydantic import Field, validator

from vizro._constants import ACCORDION_DEFAULT_TITLE
from vizro.models import VizroBaseModel
from vizro.models._models_utils import _log_call
from vizro.models._navigation._navigation_utils import _validate_pages


# TODO: if and when made public, consider naming as NavAccordion to be consistent with other
#  navigation models.
class Accordion(VizroBaseModel):
    """Accordion to be used as nav_selector in [`Navigation`][vizro.models.Navigation].

    Args:
        type (Literal["accordion"]): Defaults to `"accordion"`.
        pages (Dict[str, List[str]]): Mapping from name of a pages group to a list of page IDs. Defaults to `{}`.
    """

    type: Literal["accordion"] = "accordion"
    pages: Dict[str, List[str]] = Field({}, description="Mapping from name of a pages group to a list of page IDs.")

    _validate_pages = validator("pages", allow_reuse=True)(_validate_pages)

    @validator("pages", pre=True)
    def coerce_pages_type(cls, pages):
        if isinstance(pages, Mapping):
            return pages
        return {ACCORDION_DEFAULT_TITLE: pages}

    @_log_call
    def build(self, *, active_page_id=None):
        # Note build does not return _NavBuildType but just a single html.Div with id="nav-panel".
        # Hide navigation panel if there is only one page
        if len(list(itertools.chain(*self.pages.values()))) == 1:
            return html.Div(hidden=True, id="nav-panel")

        accordion_items = []
        for page_group, page_members in self.pages.items():
            accordion_buttons = self._create_accordion_buttons(pages=page_members, active_page_id=active_page_id)
            accordion_items.append(
                dbc.AccordionItem(
                    children=accordion_buttons,
                    title=page_group.upper(),
                    class_name="accordion-item-header",
                )
            )

        return html.Div(
            children=[
                dbc.Accordion(
                    id=self.id,
                    children=accordion_items,
                    class_name="accordion",
                    persistence=True,
                    persistence_type="session",
                    always_open=True,
                ),
            ],
            id="nav-panel",
        )

    def _create_accordion_buttons(self, pages, active_page_id):
        """Creates a button for each provided page that is registered."""
        accordion_buttons = []
        for page_id in pages:
            try:
                page = dash.page_registry[page_id]
            except KeyError as exc:
                raise KeyError(
                    f"Page with ID {page_id} cannot be found. Please add the page to `Dashboard.pages`"
                ) from exc
            accordion_buttons.append(
                dbc.Button(
                    children=[page["name"]],
                    key=page["relative_path"],
                    className="accordion-item-button",
                    active=page_id == active_page_id,
                    href=page["relative_path"],
                )
            )
        return accordion_buttons



"""Unit tests for vizro.models.Graph."""
import json

import plotly
import plotly.graph_objects as go
import pytest
from dash import dcc
from dash._callback_context import context_value
from dash._utils import AttributeDict

try:
    from pydantic.v1 import ValidationError
except ImportError:  # pragma: no cov
    from pydantic import ValidationError

import vizro.models as vm
import vizro.plotly.express as px
from vizro.actions._actions_utils import CallbackTriggerDict
from vizro.managers import data_manager
from vizro.models._action._action import Action


@pytest.fixture
def standard_px_chart_with_str_dataframe():
    return px.scatter(
        data_frame="gapminder",
        x="gdpPercap",
        y="lifeExp",
        size="pop",
        color="continent",
        hover_name="country",
        size_max=60,
    )


@pytest.fixture
def expected_graph():
    return dcc.Loading(
        dcc.Graph(
            id="text_graph",
            figure=go.Figure(
                layout={
                    "paper_bgcolor": "rgba(0,0,0,0)",
                    "plot_bgcolor": "rgba(0,0,0,0)",
                    "xaxis": {"visible": False},
                    "yaxis": {"visible": False},
                }
            ),
            config={
                "autosizable": True,
                "frameMargins": 0,
                "responsive": True,
            },
            className="chart_container",
        ),
        color="grey",
        parent_className="loading-container",
    )


class TestDunderMethodsGraph:
    def test_create_graph_mandatory_only(self, standard_px_chart):
        graph = vm.Graph(figure=standard_px_chart)

        assert hasattr(graph, "id")
        assert graph.type == "graph"
        assert graph.figure == standard_px_chart._captured_callable
        assert graph.actions == []

    @pytest.mark.parametrize("id", ["id_1", "id_2"])
    def test_create_graph_mandatory_and_optional(self, standard_px_chart, id):
        graph = vm.Graph(figure=standard_px_chart, id=id, actions=[])

        assert graph.id == id
        assert graph.type == "graph"
        assert graph.figure == standard_px_chart._captured_callable

    def test_mandatory_figure_missing(self):
        with pytest.raises(ValidationError, match="field required"):
            vm.Graph()

    def test_failed_graph_with_wrong_figure(self, standard_go_chart):
        with pytest.raises(ValidationError, match="must provide a valid CapturedCallable object"):
            vm.Graph(figure=standard_go_chart)

    def test_getitem_known_args(self, standard_px_chart):
        graph = vm.Graph(figure=standard_px_chart)

        assert graph["x"] == "gdpPercap"
        assert graph["type"] == "graph"

    def test_getitem_unknown_args(self, standard_px_chart):
        graph = vm.Graph(figure=standard_px_chart)
        with pytest.raises(KeyError):
            graph["unknown_args"]

    @pytest.mark.parametrize("title, expected", [(None, 24), ("Test", None)])
    def test_title_margin_adjustment(self, gapminder, title, expected):
        graph = vm.Graph(figure=px.bar(data_frame=gapminder, x="year", y="pop", title=title)).__call__()

        assert graph.layout.margin.t == expected
        assert graph.layout.template.layout.margin.t == 64
        assert graph.layout.template.layout.margin.l == 80
        assert graph.layout.template.layout.margin.b == 64
        assert graph.layout.template.layout.margin.r == 12

    def test_update_theme_outside_callback(self, standard_px_chart):
        graph = vm.Graph(figure=standard_px_chart).__call__()
        assert graph == standard_px_chart.update_layout(margin_t=24, template="vizro_dark")

    @pytest.mark.parametrize("template", ["vizro_dark", "vizro_light"])
    def test_update_theme_inside_callback(self, standard_px_chart, template):
        mock_callback_context = {
            "args_grouping": {
                "external": {
                    "theme_selector": CallbackTriggerDict(
                        id="theme_selector",
                        property="on",
                        value=template == "vizro_dark",
                        str_id="theme_selector",
                        triggered=False,
                    )
                }
            }
        }
        context_value.set(AttributeDict(**mock_callback_context))
        graph = vm.Graph(figure=standard_px_chart).__call__()
        assert graph == standard_px_chart.update_layout(margin_t=24, template=template)

    def test_set_action_via_validator(self, standard_px_chart, identity_action_function):
        graph = vm.Graph(figure=standard_px_chart, actions=[Action(function=identity_action_function())])
        actions_chain = graph.actions[0]
        assert actions_chain.trigger.component_property == "clickData"


class TestProcessFigureDataFrame:
    def test_process_figure_data_frame_str_df(self, standard_px_chart_with_str_dataframe, gapminder):
        data_manager["gapminder"] = gapminder
        graph_with_str_df = vm.Graph(id="text_graph", figure=standard_px_chart_with_str_dataframe)
        assert data_manager._get_component_data("text_graph").equals(gapminder)
        assert graph_with_str_df["data_frame"] == "gapminder"

    def test_process_figure_data_frame_df(self, standard_px_chart, gapminder):
        graph_with_df = vm.Graph(id="text_graph", figure=standard_px_chart)
        assert data_manager._get_component_data("text_graph").equals(gapminder)
        with pytest.raises(KeyError, match="'data_frame'"):
            graph_with_df.figure["data_frame"]


class TestBuild:
    def test_graph_build(self, standard_px_chart, expected_graph):
        graph = vm.Graph(id="text_graph", figure=standard_px_chart)

        result = json.loads(json.dumps(graph.build(), cls=plotly.utils.PlotlyJSONEncoder))
        expected = json.loads(json.dumps(expected_graph, cls=plotly.utils.PlotlyJSONEncoder))
        assert result == expected



import plotly.express as px
import pytest
from dash._callback_context import context_value
from dash._utils import AttributeDict

import vizro.models as vm
from vizro.actions import filter_interaction
from vizro.actions._actions_utils import (
    CallbackTriggerDict,
)
from vizro.managers import model_manager


@pytest.fixture
def callback_context_filter_interaction(request):
    """Mock dash.callback_context that represents a click on a continent data-point and table selected cell."""
    continent_filter_interaction, country_table_filter_interaction = request.param

    args_grouping_filter_interaction = []
    if continent_filter_interaction:
        args_grouping_filter_interaction.append(
            {
                "clickData": CallbackTriggerDict(
                    id="box_chart",
                    property="clickData",
                    value={
                        "points": [
                            {
                                "customdata": [continent_filter_interaction],
                            }
                        ]
                    },
                    str_id="box_chart",
                    triggered=False,
                )
            }
        )
    if country_table_filter_interaction:
        args_grouping_filter_interaction.append(
            {
                "active_cell": CallbackTriggerDict(
                    id="underlying_table_id",
                    property="active_cell",
                    value={"row": 0, "column": 0, "column_id": "country"},
                    str_id="underlying_table_id",
                    triggered=False,
                ),
                "derived_viewport_data": CallbackTriggerDict(
                    id="underlying_table_id",
                    property="derived_viewport_data",
                    value=[
                        {
                            "country": country_table_filter_interaction,
                            "continent": "Africa",
                            "year": 2007,
                        },
                        {
                            "country": "Egypt",
                            "continent": "Africa",
                            "year": 2007,
                        },
                    ],
                    str_id="underlying_table_id",
                    triggered=False,
                ),
            }
        )

    mock_callback_context = {
        "args_grouping": {
            "external": {
                "filters": [],
                "filter_interaction": args_grouping_filter_interaction,
                "parameters": [],
                "theme_selector": CallbackTriggerDict(
                    id="theme_selector",
                    property="on",
                    value=True,
                    str_id="theme_selector",
                    triggered=False,
                ),
            }
        }
    }
    context_value.set(AttributeDict(**mock_callback_context))
    return context_value


@pytest.fixture
def target_scatter_filtered_continent(request, gapminder_2007, scatter_params):
    continent, country = request.param

    data = gapminder_2007
    if continent:
        data = data[data["continent"].isin([continent])]
    if country:
        data = data[data["country"].isin([country])]

    return px.scatter(data, **scatter_params).update_layout(margin_t=24)


@pytest.fixture
def target_box_filtered_continent(request, gapminder_2007, box_params):
    continent, country = request.param

    data = gapminder_2007
    if continent:
        data = data[data["continent"].isin([continent])]
    if country:
        data = data[data["country"].isin([country])]

    return px.box(data, **box_params).update_layout(margin_t=24)


@pytest.mark.usefixtures("managers_one_page_two_graphs_one_table_one_button")
class TestFilterInteraction:
    @pytest.mark.parametrize("callback_context_filter_interaction", [("Africa", None), ("Europe", None)], indirect=True)
    def test_filter_interaction_without_targets_temporary_behavior(  # temporary fix, see below test
        self,
        callback_context_filter_interaction,
    ):
        # Add action to relevant component - here component[0] is the source_chart
        model_manager["box_chart"].actions = [vm.Action(id="test_action", function=filter_interaction())]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {}

        assert result == expected

    @pytest.mark.xfail  # This is the desired behavior, ie when no target is provided, then all charts filtered
    @pytest.mark.parametrize(
        "callback_context_filter_interaction,target_scatter_filtered_continent,target_box_filtered_continent",
        [
            (("Africa", None), ("Africa", None), ("Africa", None)),
            (("Europe", None), ("Europe", None), ("Europe", None)),
            (("Americas", None), ("Americas", None), ("Americas", None)),
        ],
        indirect=True,
    )
    def test_filter_interaction_without_targets_desired_behavior(
        self,
        callback_context_filter_interaction,
        target_scatter_filtered_continent,
        target_box_filtered_continent,
    ):
        # Add action to relevant component - here component[0] is the source_chart
        model_manager["box_chart"].actions = [vm.Action(id="test_action", function=filter_interaction())]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent,
            "box_chart": target_box_filtered_continent,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_filter_interaction,target_scatter_filtered_continent",
        [
            (("Africa", None), ("Africa", None)),
            (("Europe", None), ("Europe", None)),
            (("Americas", None), ("Americas", None)),
        ],
        indirect=True,
    )
    def test_filter_interaction_with_one_target(
        self,
        callback_context_filter_interaction,
        target_scatter_filtered_continent,
    ):
        # Add action to relevant component - here component[0] is the source_chart
        model_manager["box_chart"].actions = [
            vm.Action(id="test_action", function=filter_interaction(targets=["scatter_chart"]))
        ]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_filter_interaction,target_scatter_filtered_continent,target_box_filtered_continent",
        [
            (("Africa", None), ("Africa", None), ("Africa", None)),
            (("Europe", None), ("Europe", None), ("Europe", None)),
            (("Americas", None), ("Americas", None), ("Americas", None)),
        ],
        indirect=True,
    )
    def test_filter_interaction_with_two_target(
        self,
        callback_context_filter_interaction,
        target_scatter_filtered_continent,
        target_box_filtered_continent,
    ):
        # Add action to relevant component - here component[0] is the source_chart
        model_manager["box_chart"].actions = [
            vm.Action(id="test_action", function=filter_interaction(targets=["scatter_chart", "box_chart"]))
        ]

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent,
            "box_chart": target_box_filtered_continent,
        }

        assert result == expected

    @pytest.mark.xfail  # This (or similar code) should raise a Value/Validation error explaining next steps
    @pytest.mark.parametrize("target", ["scatter_chart", ["scatter_chart"]])
    @pytest.mark.parametrize("callback_context_filter_interaction", [("Africa", None), ("Europe", None)], indirect=True)
    def test_filter_interaction_with_invalid_targets(
        self,
        target,
        callback_context_filter_interaction,
    ):
        with pytest.raises(ValueError, match="Target invalid_target not found in model_manager."):
            # Add action to relevant component - here component[0] is the source_chart
            model_manager["box_chart"].actions = [
                vm.Action(id="test_action", function=filter_interaction(targets=target))
            ]

    @pytest.mark.parametrize(
        "callback_context_filter_interaction,target_scatter_filtered_continent",
        [
            ((None, "Algeria"), (None, "Algeria")),
            ((None, "Albania"), (None, "Albania")),
            ((None, "Argentina"), (None, "Argentina")),
        ],
        indirect=True,
    )
    def test_table_filter_interaction_with_one_target(
        self,
        callback_context_filter_interaction,
        target_scatter_filtered_continent,
    ):
        model_manager["box_chart"].actions = [
            vm.Action(id="test_action", function=filter_interaction(targets=["scatter_chart"]))
        ]

        model_manager["vizro_table"].actions = [vm.Action(function=filter_interaction(targets=["scatter_chart"]))]
        model_manager["vizro_table"].pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_filter_interaction, target_scatter_filtered_continent, target_box_filtered_continent",
        [
            ((None, "Algeria"), (None, "Algeria"), (None, "Algeria")),
            ((None, "Albania"), (None, "Albania"), (None, "Albania")),
            ((None, "Argentina"), (None, "Argentina"), (None, "Argentina")),
        ],
        indirect=True,
    )
    def test_table_filter_interaction_with_two_targets(
        self,
        callback_context_filter_interaction,
        target_scatter_filtered_continent,
        target_box_filtered_continent,
    ):
        model_manager["box_chart"].actions = [
            vm.Action(id="test_action", function=filter_interaction(targets=["scatter_chart", "box_chart"]))
        ]

        model_manager["vizro_table"].actions = [
            vm.Action(function=filter_interaction(targets=["scatter_chart", "box_chart"]))
        ]
        model_manager["vizro_table"].pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent,
            "box_chart": target_box_filtered_continent,
        }

        assert result == expected

    @pytest.mark.parametrize(
        "callback_context_filter_interaction, target_scatter_filtered_continent, target_box_filtered_continent",
        [
            (("Africa", "Algeria"), ("Africa", "Algeria"), ("Africa", "Algeria")),
            (("Europe", "Albania"), ("Europe", "Albania"), ("Europe", "Albania")),
            (("Americas", "Argentina"), ("Americas", "Argentina"), ("Americas", "Argentina")),
        ],
        indirect=True,
    )
    def test_mixed_chart_and_table_filter_interaction_with_two_targets(
        self,
        callback_context_filter_interaction,
        target_scatter_filtered_continent,
        target_box_filtered_continent,
    ):
        model_manager["box_chart"].actions = [
            vm.Action(id="test_action", function=filter_interaction(targets=["scatter_chart", "box_chart"]))
        ]

        model_manager["vizro_table"].actions = [
            vm.Action(function=filter_interaction(targets=["scatter_chart", "box_chart"]))
        ]
        model_manager["vizro_table"].pre_build()

        # Run action by picking the above added action function and executing it with ()
        result = model_manager["test_action"].function()
        expected = {
            "scatter_chart": target_scatter_filtered_continent,
            "box_chart": target_box_filtered_continent,
        }

        assert result == expected

    # TODO: Simplify parametrization, such that we have less repetitive code
    # TODO: Eliminate above xfails
    # TODO: Complement tests above with backend tests (currently the targets are also taken from model_manager!
    # see _get_action_callback_mapping eg.)



import json

import dash.development
import plotly

STRIP_ALL = object()


def _strip_keys(object, keys):
    """Strips from a JSON object all entries where the key is in keys, regardless of how deeply it's nested."""
    if isinstance(object, dict):
        object = {key: _strip_keys(value, keys) for key, value in object.items() if key not in keys}
    elif isinstance(object, list):
        object = [_strip_keys(item, keys) for item in object]
    return object


def _component_to_dict(component: dash.development.base_component.Component) -> dict:
    """Prepares a Dash component for comparison by conversion to JSON object."""
    return json.loads(json.dumps(component, cls=plotly.utils.PlotlyJSONEncoder))


def assert_component_equal(left, right, *, keys_to_strip=None):
    """Checks that the left and right Dash components are equal, ignoring keys_to_strip.

    If keys_to_strip is set to STRIP_ALL then only the type and namespace of component
    will be compared, similar to doing isinstance.

    Examples:
        >>> from dash import html
        >>> assert_component_equal(html.Div(), html.Div())
        >>> assert_component_equal(html.Div(id="a"), html.Div(), keys_to_strip={"id"})
        >>> assert_component_equal(html.Div([html.P(), html.P()], id="a"), html.Div(id="a"), keys_to_strip={"children"})
        >>> assert_component_equal(html.Div(html.P(), className="blah", id="a"), html.Div(), keys_to_strip=STRIP_ALL)
    """
    keys_to_strip = keys_to_strip or {}
    if keys_to_strip is STRIP_ALL:
        # Remove all properties from the component dictionary, leaving just "type" and "namespace" behind.
        keys_to_strip = {"props"}

    left = _strip_keys(_component_to_dict(left), keys_to_strip)
    right = _strip_keys(_component_to_dict(right), keys_to_strip)
    assert left == right
